#include "array.h"
#include <stdlib.h>
#define ALLOC_CHUNK_SIZE 32

/**
 * \brief Increases the array's available memory size.
 * This function is mainly for internal use. It is
 * called implicitly bu array_add().
 *
 * \param array An array.
 * \param new_places Number of new places to add.
 */
int array_increase(Array *array, int new_places) 
{
    int newTotalSize;
    void **newElements;
    int success = 0;
    int i;

    if (new_places > 0)
    {
	newTotalSize = array->max_length + new_places;
	newElements = (void **)malloc(newTotalSize *sizeof(void *));
	if (newElements > 0 ) {
	    for (i=0;i<array->length;i++) {
		newElements[i] = array->elements[i];
	    }
	    free(array->elements);
	    array->elements = newElements;
	    array->max_length = newTotalSize;
	    success = 1;
	}
    }
    return success;
}

/**
 * \brief Creates a new array.
 * Creates a new array with a buffer for 32 elements.
 */
Array *array_create()
{
    return array_create_sized(ALLOC_CHUNK_SIZE);
}

/**
 * \brief Creates a new array.
 * \param bufferSize Maximal number of contained elements.
 * Creates a new array with a pre-set buffer size.
 * Prevents re-allocating memory space if the array
 * size is known in advance.
 */
Array *array_create_sized(int bufferSize) 
{
    Array *result = NULL;
    Array *newArray;
    
    if (bufferSize < 2) bufferSize = 2;

    newArray = (Array *)malloc(sizeof(Array));
    if (newArray != NULL)
    {
	newArray->elements = (void **)malloc(bufferSize * sizeof(void *));
	if (newArray->elements == NULL)
	{
	    free(newArray);
	    result = NULL;
	}
	else
	{
	    newArray->max_length = bufferSize;
	    newArray->length = 0;
	    result = newArray;
	}
    }
    else
    {
	result = NULL;
    }
    return result;
}

/**
 * \brief Appends an element to an array.
 * If necessary, the array will be increased automatically.
 */
int array_append(Array *array, void *newElement) 
{
    int success = 0; /* pessimistic */
    
    if (array->length >= array->max_length)
    {
	/* No Free space; redim */
	if (array_increase(array, ALLOC_CHUNK_SIZE)) {
	    /* redim was successful */
	    success = array_append(array, newElement);
	}
	
    } else {
	
	array->elements[array->length] = newElement;
	array->length++;
	success = 1;
    }
    
    return success;
    
}

/**
 * \brief Removes an element from an array.
 * \param array An array.
 * \param index Index of the element.
 * \param destructor Destructor function which has to be called to erase the element.
 */
void array_remove(Array *array, int index, DestructorFnc destructor) 
{
    int i;
    
    if ((index >=0) && (index < array->length))
    {
	if (destructor != NULL) {
	    destructor(array->elements[index]);
	}
	for (i=index+1; i<array->length; i++) {
	    array->elements[i-1] = array->elements[i];
	}
	array->length--;
    }
}


/**
 * \brief Removes an element from an array.
 * \param array An array.
 * \param element An element.
 * \param destructor Destructor function which has to be called to erase the element.
 */
void array_remove_element(Array *array, void *element, DestructorFnc destructor) 
{
    int index;
    index = array_find_index(array, element);
    if (index >=0) {
	array_remove(array,index, destructor);
    }
}

/**
 * \brief Finds the index of a specified element.
 * \param array An array.
 * \param element An element to find.
 * \return Index on success, else -1.
 */
int array_find_index(Array *array, void *element) 
{
    int i;
    int result = -1;

    for(i=0; i<array->length; i++)
    {
	if (array->elements[i] == element) {
	    result = i;
	    break;
	}
    }
    return result;
}

/**
 * \brief Finds an array element.
 * \param array An array.
 * \param equals Comparator function. Compares each element to user_data. Must return 0 on failure and 1 on success.
 * \param user_data Data to be compared.
 *
 * Finds the first element for which equals(element, user_data) != 0.
 *
 * \return Pointer to an element on success, NULL on failure.
 * \see EqualsFnc
 */
void *array_find_equals(Array *array, EqualsFnc equals, void * user_data) 
{
    int i;
    void *result = NULL;

    for(i=0; i<array->length; i++)
    {
	if (equals(array->elements[i], user_data)) {
	    result = array->elements[i];
	    break;
	}
    }
    return result;
}


/**
 * \brief Inserts an element before the specified index.
 * \param array An array.
 * \param index The index.
 * \param newElement Element to be inserted.
 *
 * \return true (1) if the element was inserted, false (0) else.
 */
int array_insert_before(Array *array, int index, void *newElement)
{

    int success = 0;			/* pessimistic */
    int i;
    
    if (array->length >= array->max_length)
    {
	/* insufficient memory */
	if (array_increase(array, ALLOC_CHUNK_SIZE))
	{
	    success = array_insert_before(array,index,newElement);
	} 
    }
    else /* array has enough memory */
    {
	if (index < 0) index = 0;
	if (index >= array->length) index = array->length;

	/* shift elements right */
	for (i=array->length-1; i>=index; i--) {
	    array->elements[i+1] = array->elements[i];
	}

	/* insert element */
	array->elements[index] = newElement;
	array->length ++;
	success = 1; /* success */
    }
    
    return success;
    
}

/**
 * \brief Returns an element at a specified index.
 * \param array An array.
 * \param index An index
 */
void *array_get(Array *array, int index) 
{
    if (index < array->length) {
	return array->elements[index];
    } else {
	return NULL;
    }
}


/**
 * \brief Removes all elements.
 * \param array An array
 * \param destructor A function which frees each element and its contents.
 * The destructor is responsible for de-allocating the element
 * and all its contents. If the destructor is NULL, the element data
 * will not be freed.
 */ 
void array_clear(Array *array, DestructorFnc destructor)
{
    int i;
    if (destructor != NULL) {
	/* destroy elements */
	for (i=0; i<array->length; i++) {
	    destructor(array->elements[i]);
	}
    }
    array->length = 0;
}

/**
 *
 * \brief Removes the array and all elements.
 * \param array An array
 * \param destructor A function which frees each element and its contents.
 * The destructor is responsible for de-allocating the element
 * and all its contents. If the destructor is NULL, the element data
 * will not be freed.
 */ 
void array_destroy(Array *array, DestructorFnc destructor)
{
    array_clear(array, destructor);
    free(array->elements);
    free(array);
}

