/* -*-C-*- */
#ifndef ARRAY_H
#define ARRAY_H


/**
 * \file array.h
 * \brief Dynamic Arrays for arbitrary data.
 */

/**
 * \brief Extracts an element at a specified position.
 * This is a macro alternative to array_get. It is faster,
 * but does not check boundaries.
 */
#define array_at(array, index) ((array)->elements[index]) 


/**
 * \brief Dynamic Array.
 * An Array contains an array of void pointers which is
 * allocated dynamically and automatically resized on
 * demand. References to elements are stored as void pointers
 * in that array.
 * 
 */
typedef struct 
{
    void **elements;			/**< Array of elements */
    int length;				/**< Actual number of stored elements */
    int max_length;			/**< Maximal number of elements */
} Array;


/**
 * \brief Destructor function.
 * \param element element to be destroyed.
 *
 * A destructor function frees the elements data
 * and the element itself. It is used in array functions
 * which remove elements.
 */
typedef void (* DestructorFnc) (void *element);

/**
 * \brief Equivalence comparator.
 * \param element Element to be compared.
 * \param used_data Data used for comparison.
 *
 * An EqualsFnc returns 1 if a given element
 * is "equal" to the user data, or 0 it it is not.
 * This type of function can be used to compare certain
 * attributes of a given element with supplied data.
 *
 * \see array_find_equals()
 */
typedef int (* EqualsFnc) (void *element, void *user_data);

int array_increase(Array *array, int nNewEntries);
Array *array_create();
Array *array_create_sized(int bufferSize) ;
int array_append(Array *array, void *newElement) ;
void array_remove(Array *array, int index, DestructorFnc destructor) ;
void array_remove_element(Array *array, void *element, DestructorFnc destructor) ;
int array_find_index(Array *array, void *element) ;
void *array_find_equals(Array *array, EqualsFnc equals, void *user_data) ;

int array_insert_before(Array *array, int index, void *element);
void *array_get(Array *array, int index) ;
void array_clear(Array *array, DestructorFnc destructor);
void array_destroy(Array *array, DestructorFnc destructor);



#endif
