#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include "sockets.h"
#include <fcntl.h>
#include "sys_util.h"
#include "string_util.h"
#include <ctype.h>
#include "log.h"

/**
  \file autoproc.c
  \brief The command-line client for autoprocd.
*/

#define LOCAL_SOCKET "/tmp/autoproc-0"	/**< FIFO path */
#define HOSTNAME_SIZE 256		/**< Size of m_hostname */
static char m_hostname[HOSTNAME_SIZE];	/**< Hostname which is displayed in the prompt  */
static int m_port;			/**< Remote port used for a network connection */
static int m_is_remote_connection;	/**< true (1) if the current connection is a remote connection */
static int m_is_interactive;		/**< true (1) if this program runs in interactive mode  */

static void autoproc_set_host_port(const char *optval);
static void autoproc_set_host(const char *command);
static void autoproc_print_prompt();
static void autoproc_send_command(char *cmd, FILE *stream);
static void autoproc_open_connection();
static void autoproc_close_connection();
static void autoproc_print_prompt();

static FILE* m_current_stream = NULL;	/**< I/O stream on curent socket  */
int m_current_socket = -1; /**< socket used for communication with autoprocd  */

#define UNIX_FIFO_NAME "/tmp/autoproc-0" /**< Path of UNIX FIFO  */

/**
 * \brief Opens the current socket and stream.
 */
static void autoproc_open_connection() 
{
    if (m_is_remote_connection)
    {
	m_current_socket = sockets_inet_connect(m_hostname, m_port);
    }
    else
    {
	m_current_socket = sockets_unix_connect(UNIX_FIFO_NAME, 0);
    }
    if (m_current_socket < 0)
    {
	log_perror("Opening socket");
	m_current_stream = NULL;
	strcpy(m_hostname, "error");
    }
    else
    {
	m_current_stream = fdopen(m_current_socket, "a+");
	if (m_current_stream == NULL) {
	    log_perror("opening stream");
	    strcpy(m_hostname, "error");
	}
    }
}

/**
 * \brief Closes the current socket and stream.
 */
static void autoproc_close_connection()
{
    if (m_current_stream != NULL)
    {
	fclose(m_current_stream);
	m_current_stream = NULL;
    }
    if (m_current_socket != 0)
    {
	close(m_current_socket);
	m_current_socket = 0;
    }
}


/**
 *\brief Extracts host and port from a string "<host>[:<port>]"
 */
static void autoproc_set_host_port(const char *optval)
{
    const char *ptr;
    char *pptr;
    int brk;

    brk = 0;

    ptr = optval;
    
    /* first non-whitespace character points to the hostname */
    while((!brk) && (isspace(*ptr)))
    {
	ptr++;
	brk = (*ptr == 0);
    }

    if (!brk)
    {
	pptr = strchr(ptr, ':');
	if (pptr != NULL)
	{
	    *pptr = 0;
	    pptr++;
	    m_port = atoi(pptr);
	}
	else
	{
	    m_port = 3510;
	}

	string_util_strncpy(m_hostname, ptr, HOSTNAME_SIZE);
	string_util_trim(m_hostname);
    }
    else
    {
	m_hostname[0] = 0;
	m_port = 0;
    }
};

/**
 * \brief Treats input line which starts with "host"
 * \param command An input line.
 */
static void autoproc_set_host(const char *command)
{
    const char *ptr;
    
    ptr = command;

    /* Find first separator (remove "host") */
    while((*ptr !=0 ) && (!isspace(*ptr)))
    {
	ptr++;
    }

    if (*ptr != 0) autoproc_set_host_port(ptr);

    m_is_remote_connection = ((m_port != 0)
			      && (strcmp(m_hostname, "localhost")));
}

/**
 * \brief Prints out a prompt.
 */
static void autoproc_print_prompt()
{
     if (m_is_interactive)
     {
	 if(m_is_remote_connection)
	     printf("(autoproc:%s) ", m_hostname);
	 else
	     printf("(autoproc:local) ");

	 fflush(stdout); 
     }
}

#define INPUT_BUFFER_SIZE 128		/**< Size of input buffer for autoproc_set_command */

/**
 * \brief Sends a command.
 */
static void autoproc_send_command(char *cmd, FILE *stream)
{
    static char input_buffer[INPUT_BUFFER_SIZE];
    char *rc;

    if (stream == NULL) {
	log_error("Not connected!");
	return;
    }
    
    /* propagate command */
    fputs(cmd, stream);
    fflush(stream);

    /* print reply */
    while((rc = fgets(input_buffer, INPUT_BUFFER_SIZE -1, stream))
	  != NULL)

    {
	if (input_buffer[0] != '.')
	{
	    printf( "%s", input_buffer);
	    fflush(stdout); 
	}
	else
	{
	    break;
	}
    }
    
    if (rc == NULL)
    {
        log_perror("Sending command");
    }
}


#define LINE_BUFFER_SIZE 1024		/**< Size of line buffer for main */
/**
 * \brief Main function.
 */
int main(int argc, char **argv)
{
    int i;
    char *rc;
    int n_descriptors;
    int option_char;
    fd_set fdset;
    char line_buffer[LINE_BUFFER_SIZE];
    char username[80];
    char *passwd;
    
    
    m_is_interactive = 0;
    m_is_remote_connection = 0;
    m_hostname[0] = 0;
	
    /* initialize options */
    username[0] = 0;

    m_current_socket = 0;
    m_current_stream = NULL;


    /* Initialize logging before configuration. */
    log_initialize("autoproc");
    log_add_handler(log_console_handler, LOG_LEVEL_WARNING);
    
    /* look for command line options */
    while( (option_char=getopt(argc,argv,"u:h:")) != -1)
    {
	switch(option_char)
	{
	case 'h':
	    autoproc_set_host_port(optarg);
	    m_is_remote_connection = (m_port != 0);
	    autoproc_print_prompt();
	    break;
			    
	default:
	    log_fatal("Syntax: autoproc [ -h <host> ].");
	    return 1;
	}
    }

    if (argc > optind)			/* Non-interactive mode */
    {
	m_is_interactive = 0;
	
	autoproc_open_connection();
	if ((m_current_socket < 0) || (m_current_stream == NULL))
	{
	    log_fatal("Cannot connect to autoprocd");
	    exit(1);
	}
	
	/* concatenate arguments */
	line_buffer[0] = 0;
	for(i=optind;i<argc;i++)
	{
	    if (strlen(line_buffer) + strlen(argv[i]) < LINE_BUFFER_SIZE-2)
	    {
		if (i>1)
		    strcat(line_buffer, " ");
		strcat(line_buffer, argv[i]);
	    }
	}

	/* write command to socket */
	fprintf(m_current_stream, "%s\n", line_buffer);

	/* wait for reply */
	while((!feof(m_current_stream))
	      && ((rc = fgets(line_buffer, LINE_BUFFER_SIZE - 1, m_current_stream))
		  != NULL)
	      && (line_buffer[0] != '.'))
	{
	    printf( "%s", line_buffer);
	}
  	fflush(stdout); 
	autoproc_close_connection();
	
    } /* non-interactive */
    else 
    {					/* interactive mode */

	m_is_interactive = 1;
	
	autoproc_open_connection();
	if ((m_current_socket < 0) || (m_current_stream == NULL))
	{
	    log_fatal("Cannot connect to autoprocd.");
	    exit(1);
	}

	autoproc_print_prompt();
   	fflush(stdout); 

	while(1)
	{

	    FD_ZERO(&fdset);
	    FD_SET(0, &fdset);
	    /* FD_SET(s, &fdset); */


	    n_descriptors = select(FD_SETSIZE, &fdset, NULL, NULL, NULL);
	    if (n_descriptors > 0)
	    {
		if (FD_ISSET(0, &fdset))
		{
		    /*	    printf("test\n"); */
		    
		    /*    bytes = read(STDIN_FILENO, line_buffer, LINE_BUFFER_SIZE-1);
			  if (bytes > 0)
			  write(s, line_buffer, strlen(line_buffer) + 1);*/

		    rc = fgets(line_buffer,LINE_BUFFER_SIZE,stdin);
		    if (rc != NULL)
		    {
			if ((!strcmp(line_buffer, "q\n"))
			    || (!strcmp(line_buffer, "quit\n"))
			    || (!strcmp(line_buffer, "exit\n")))
			{
			    autoproc_close_connection();
			    exit(0);
			}
			else if (strstr(line_buffer, "host") != NULL)
			{
			    autoproc_close_connection();
			    autoproc_set_host(line_buffer);
			    autoproc_open_connection();
			    autoproc_print_prompt();
			    
			}
			else if (!strncmp(line_buffer, "login", 5))
			{
			    passwd = getpass("Password: ");
			    if (passwd != NULL)
			    {
				strcpy(line_buffer, "pw ");
				strncat(line_buffer, passwd, 20);
				strcat(line_buffer, "\n");
				autoproc_send_command(line_buffer, m_current_stream);
			    }
			    autoproc_print_prompt();
			}
			else if (!strncmp(line_buffer, "local",5))
			{
			    autoproc_close_connection();
			    m_is_remote_connection = 0;
			    autoproc_open_connection();
			    autoproc_print_prompt();
			}
			else
			{ /* server command */
			    autoproc_send_command(line_buffer, m_current_stream);
			    autoproc_print_prompt();
			} /* remote/local */
		    } /* fgets OK */
		} /* FD_ISSET stdin */
	    } /* select OK */
	} /* while */
    } /* interactive */

    autoproc_close_connection();
    
    return 0;
}







