#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <sys/wait.h>
#include <sys/io.h>

#include <pth.h>

#include "log.h"
#include "sys_util.h"
#include "process.h"
#include "setup.h"
#include "process_control.h"
#include "interpreter.h"
#include "server.h"


#define DEFAULT_PORT 3510
#define HOME_DIR_LEN 256


/**
 * \brief Main Program.
 * \author Josef Wilgen
 */


/**
 * \brief Checks if an instance of this program is already running
 *
 * This is a simple test which reads the PID file and tests
 * if a process with the found PID is running.
 */
static int autoprocd_is_running()
{

    int result = 0;
    int pid = 0;
    FILE *file;
    
    file = fopen(setup_get_configuration()->pid_file, "r");
    if (file != NULL)
    {
	if (fscanf(file, "%d",&pid) > 0)
	    result = ((pid > 0) && (!kill(pid, 0)));
	else
	    result = 0;
	fclose(file);
    }
    return result;

}


/**
 * \brief Writes the PID to a file
 */
static int autoprocd_write_pid()
{
    int result = 0;
    FILE *file;

    Configuration *config;

    log_debug("autoproc_write_pid()");
	      
    config = setup_get_configuration();
    if (config->pid_file[0] != 0) {
	file = fopen(config->pid_file, "w");
	if (file != NULL)
	{
	    result = (fprintf(file, "%d", getpid()) > 0);
	    fclose(file);
	}
	else
	{
	    log_perror("Writing PID file");
	}
    }
    else
    {
	log_debug("No PID file specified.");
    }
    
    return result;
}


/**
 * \brief Child signal Handler.
 *
 * The child signal handler frees the terminated process
 * from Zombie state and wakes the parent process from select()
 */
static void autoprocd_sig_child(int x)
{
    pid_t pid;
    int status;

    pid = wait(&status);
    log_debug("autoprocd_sig_child()");
    process_control_notify_termination(pid);
    
}


/**
 * \brief Main Program.
 *
 * Reads command line options, reads configuration file,
 * initializes, starts threads.
 */
int main(int argc, char **argv)
{

    int fork_pid;
    int return_code;
    pth_attr_t thread_attr;
    char home_dir[HOME_DIR_LEN];

    uid_t new_uid;
    gid_t new_gid;

    pth_t control_thread;
    pth_t server_thread;

    sigset_t caught_signals; /* Signals which are caught by pth_sigwait() below. */
    int signal_number;
    

    Configuration *config_options;
    Options *options;

    new_uid = getuid();
    new_gid = getgid();

    options = options_create_from_args(argc, argv);

    /* Initialize syslog in advance - may not be needed */
    openlog("autoprocd", 0, LOG_DAEMON);

    /* Basic logging - before setup */
    log_initialize("autoprocd");  
    log_add_handler(log_console_handler, LOG_LEVEL_ERROR);

    /* Fork if daemon */
    if (!options->set_debug_enabled)
    {
	fork_pid = fork();
	if (fork_pid > 0)			/* exit parent process */
	{
	    exit(0);
	}
	else if (fork_pid < 0)			/* error */
	{
	    log_fatal("Fork: %s", strerror(errno));
	    exit(1);
	}
	/* child process */

	setsid();
	/* chdir("/"); */
	/* umask(0);			 Only the owner has access */
    }

    /* If the option -u is set, the configuration file
     * is expected to be in
     * the user's home directory. */
    if (options->set_user) {

	/* Determine home directory */
	if(!sys_util_findUID(options->user_name, &new_uid, &new_gid)) {
	    log_fatal("autoproc: Unknown user %s", options->user_name);
	    exit(1);
	}
	if (new_uid < 0)
	{		
	    log_fatal("User %s not found.", options->user_name);
	    return(1);
	}
    }
    
    /* Find the configuration file */
    if (!options->set_config_file)	/* If not given as an argument */
    {
	if (new_uid == 0)
	{
	    /* The default */
	    strncat(options->config_file, "/etc/autoproc.conf", OPTIONS_LEN_CONFIG_FILE);
	}
	else
	{
	    /* Look for autoproc.conf in user's home directory */
	    if (sys_util_findHomeDir(new_uid, options->config_file, OPTIONS_LEN_CONFIG_FILE))
		strncat(options->config_file, "/autoproc.conf", OPTIONS_LEN_CONFIG_FILE);
	    else
	    {
		log_fatal("Home directory of UID %d cannot be found.", new_uid);
		return(1);
	    }
	}
    }

    /* Read the configuration file */
    config_options = setup_initialize(options, options->config_file);

    /* Logging is properly initialized after setup_initialize */
    log_info("Starting autoprocd");

    /* Open I/O Ports
     * Some processes with non-root UID like to access I/O-Ports.
     * Therefore autoprocd must be run as root wit the -I option.
     */ 
    if (options->set_ioperm) {
	log_info("Setting port I/O permissions.");
	return_code = ioperm(0x0, 0x3ff,1);
	if (return_code != 0) {
	    log_warning("Setting port I/O permissions: %s", strerror(errno));
	}
    }

    /* Check if autoprocd is already running (needs PID file)
     * This is only a simple test.
     * To ensure that only one autoprocd is running on the system,
     * the running processes would have to be investigated.
     */
    if (autoprocd_is_running())
    {
	log_fatal("Autoprocd is already running!");
	exit(1);
    }

    /* Write PID to file.
     * The PID file is used to stop autoprocd via init script.
     */
    if (!autoprocd_write_pid())
    {
	log_fatal("Can't write to %s", config_options->pid_file);
	exit(1);
    }

    /* The -u command line option overrides the configuration.
     * If no user is set on the command line or if autoprocd is
     * called with -u root, the values from the configuration
     * files are used. Take UID/GID from configuration.
     */
    if (new_uid == 0)
    {
	new_uid = config_options->uid;
	new_gid = config_options->gid;
    }
    
    /* Change GID */
    if ((new_gid != 0) && (new_gid != getgid()))
    {
	if (setgid(new_gid) != 0)
	{
	    log_fatal("Can't set Group ID");
	    return(1);
	}
    }
    
    /* Change UID */
    if ((new_uid != 0) && (new_uid != getuid()))
    {
	if (setuid(new_uid) != 0)
	{
	    log_fatal("Can't set User ID");
	    return(1);
	}
	else
	{
	    if (sys_util_findHomeDir(config_options->uid, home_dir, HOME_DIR_LEN))
		setenv("HOME", home_dir, 1);
	}
    }

    setup_execute_global_init();
    
    /* Initialize pth library */
    pth_init();

    /* Initialize process control (must be done after pth_init)! */
    process_control_initialize(setup_get_processes());

    /* Install signal handlers */
    signal(SIGCHLD, autoprocd_sig_child);
    signal(SIGPIPE, SIG_IGN);

    /* These signals are handled in pth_sigwait() (see below) */
    sigemptyset(&caught_signals);
    sigaddset(&caught_signals, SIGTERM);
    sigaddset(&caught_signals, SIGINT);
    sigaddset(&caught_signals, SIGQUIT);
    sigaddset(&caught_signals, SIGHUP);
    
    /* Start Processes */
    log_info("Starting processes.");
    process_control_start_all_processes(1);

    /* Create Threads */
    thread_attr = pth_attr_new();
    pth_attr_set(thread_attr, PTH_ATTR_JOINABLE, FALSE);
    pth_attr_set(thread_attr, PTH_ATTR_PRIO, PTH_PRIO_MAX);
    log_info("Starting process control.");
    control_thread = pth_spawn(thread_attr, process_control_update_thread, NULL);


    /* Start Server thread */
    log_info("Starting Server.");
    pth_attr_set(thread_attr, PTH_ATTR_PRIO, PTH_PRIO_STD);
    server_thread = pth_spawn(thread_attr, server_main_thread, NULL);



    pth_sigmask(SIG_BLOCK, &caught_signals, NULL);


    /* This thread waits until a signal is sent.
	After a signal is delivered, this guarantees an orderly
	shutdown of all threads and managed processes.
        It cannot be used for signals which require an
	immediate shutdown. */
    pth_sigwait(&caught_signals, &signal_number);

    log_info("SIGNAL %d received. Shutting down.", signal_number);
    pth_cancel(server_thread);
    pth_cancel(control_thread);
    pth_join(server_thread, NULL);
    pth_join(control_thread, NULL);

    process_control_shutdown_all_processes();
    server_shutdown();

    log_info("All processes cancelled. Closing autoprocd.");

    return 0;
    
}







