#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "string_util.h"
#include "config_parser.h"


#define CFG_LEN_BUFFER 256
static char line_buffer[CFG_LEN_BUFFER];


#define LINE_EOF -1			/* End of file reached. */
#define LINE_COMMENT 0			/* This line is a comment */
#define LINE_ASSIGNMENT 1		/* This line is an assignment */
#define LINE_STARTBLOCK 2		/* This line starts of a block */
#define LINE_ENDBLOCK 3			/* This line ends a block */
#define LINE_INVALID 4			/* This line is invalid. */

static int config_parser_parse_line(FILE *f, char **arg1, char **arg2);

/**
 * \brief Parses an input line.
 * \param arg1 Contents pointer 1 
 * \param arg2 Contents pointer 2
 *
 * This function reads an input line and finds out if it is
 * the beginning or end of a block or if it is an assignment.
 * In arg1 and arg2, it returns two pointers to contents
 * of the line which have context-dependant meanings.
 * If the line contains the beginning of a block, the first
 * pointer is set to the block type, the second pointer is set
 * to the block name.
 * If the line is an assignment, the first pointer is set to the
 * name, the second is set to the value.
 *
 * \return Line type. LINE_STARTBLOCK, LINE_ENDBLOCK, LINE_ASSIGNMENT, etc.
 * or LINE_EOF if the end of the file is reached.
 * 
 */
static int config_parser_parse_line(FILE *f, char **arg1, char **arg2)
{
    int rv;
    int i;
    int len;
    int found;
    char *sptr;
    
    
    if (fgets(line_buffer, CFG_LEN_BUFFER, f) != NULL)
    {
	*arg1 = NULL;
	*arg2 = NULL;
	
	line_buffer[CFG_LEN_BUFFER-1] = 0;

	/* End line with first comment sign */
	sptr = strchr(line_buffer, '#');
	if (sptr != NULL) *sptr = 0;

	/* Remove leading and trailing spaces */
	len = string_util_trim(line_buffer);

	/* strip trailing newline char */
	if (line_buffer[len-1] == '\n')
	{
	    line_buffer[len-1] = 0;
	    len--;
	}
	
	for(i=0;i<len;i++)
	{
	    if (line_buffer[i] == '#')
	    {
		line_buffer[i] = 0;
		break;
	    }
	}
	
	if ((strlen(line_buffer) == 0) || (line_buffer[0] == '#'))
	{
	    rv = LINE_COMMENT;
	}
	else if (line_buffer[len-1] == '{')
	{
	    if (len > 1)
	    {
		line_buffer[len-1] = 0;

		found = -1;
		for (i=0;((i<len-1) && (line_buffer[i] != 0));i++)
		{
		    if (isspace(line_buffer[i])) /* line_buffer[0] cannot be whitespace */
		    {
			found = i;
			break;
		    }
		}
		 
		if (found >= 0)
		{
		    *arg1 = line_buffer;
		    line_buffer[found] = 0;
		    
		    /* find first non-whitespace after *arg1 */
		    for (i=found+1,found=-1;((i<len-1) && (line_buffer[i] != 0));i++)
		    {
			if (!isspace(line_buffer[i]))
			{
			    found = i;
			    break;
			}
		    }

		    if ((found >= 0) && (found < len-1))
		    {
			*arg2 = &line_buffer[found];
			/* find first whitespace after *arg2, replace with zero */
			for(i=found; ((i<len-1) && (line_buffer[i] != 0)); i++)
			{
			    if (isspace(line_buffer[i]))
			    {
				line_buffer[i] = 0;
				break;
			    }
			}
		    }
		    
		} /* whitespace found */
	    }
	     
	    rv = LINE_STARTBLOCK;

	}
	else if (line_buffer[0] == '}')
	{
	    rv = LINE_ENDBLOCK;
	}
	else
	{				/* valid line */
	    *arg2 = strpbrk(line_buffer, ":=");
	    if (*arg2 != NULL)
	    {
		/* replace colon with 0 and separate *arg2 and *arg1 */
		**arg2 = 0;
		(*arg2)++;
		string_util_fast_trim(*arg2, arg2, 1);

		*arg1 = &line_buffer[0];
		string_util_fast_trim(*arg1, arg1, 0);

		rv = LINE_ASSIGNMENT;
	    }
	    else
	    {
		rv = LINE_INVALID;
	    }
	}
    }
    else
    {
	rv = LINE_EOF;
    }

    return rv;

}

/**
 *  \brief Read a configuration file.
 *
 *  This function parses a configiration file and calls callback
 *  functions at the beginning of each block, at the end of each block,
 *  and when an assignment occurs.
 *  
 *  \param filename Name of the configuration file
 *  \param f_start_block Callback function. Called at the beginning of a block.
 *          Arguments: Block type and block name.
 *  \param f_end_block Callback function. Called at the end of a block.
 *  \param f_set_attr Callback function. Called when an attribute is set.
 *                           Arguments: Variable name and value.
 */
int config_parser_read(const char *filename,
	     void (*f_start_block)(const char *, const char *),
	     void (*f_end_block) (),
	     void (*f_set_attr)(const char *, const char *))
{
    char *attr;
    char *value;
    int line_info;
    int rv;
    FILE *f;
    
    if ( (f = fopen(filename, "r")) != NULL)
    {
	rv = 1;
	
	while ( (line_info = config_parser_parse_line(f, &attr, &value) ) != LINE_EOF)
	{
	    switch (line_info)
	    {
	    case LINE_STARTBLOCK:
		if (f_start_block != NULL)
		    f_start_block(attr,value);
		break;
		
	    case LINE_ENDBLOCK:
		
		if (f_end_block != NULL)
		    f_end_block();
		break;
		    
	    case LINE_ASSIGNMENT:
		
		if (f_set_attr != NULL)
		    f_set_attr(attr,value);
		break;
		
	    default:
		break;
		
	    } /* switch */
	    
	} /* parse_line */
	       
	fclose(f);
	       
    } /* fopen */
    else
    {
	rv = 0;
    }

    return rv;

}


