#include <string.h>
#include <unistd.h>  /* for getuid */
#include <stdlib.h> /* setenv */
#include <stdio.h>  /* snprintf */
#include "connection.h" /* auth_modes */
#include "configuration.h"
#include "environment.h"
#include "sys_util.h"
#include "string_util.h"
#include "log.h"
#include <string.h> 			/* snprintf & co */


#define DEFAULT_PIDFILE "/var/run/autoprocd.pid"


/**
 * \brief Creates and initializes a Configuration.
 */
Configuration *configuration_create() 
{
    Configuration *result;
    Array *newEnv;

    result = (Configuration *)malloc(sizeof(Configuration));
    if (result == NULL) {
	log_error("configuration_create: Cannot allocate memory!");
	return NULL;
    }

    newEnv = array_create();
    if (newEnv == NULL) {
	log_error("configuration_create: Cannot allocate memory!");
	free(result);
	return NULL;
    }

    result->environment = newEnv;
    configuration_set_defaults(result);
    return result;
    
}


/**
 * \brief Clears a configuration.
 * Clears the environment array.
 */
void configuration_clear(Configuration *conf) 
{
    array_clear(conf->environment, environment_destroy);
    configuration_set_defaults(conf);
}


/**
 * \brief Destroys a Configuration.
 * Frees a Configuration and its data.
 */
void configuration_destroy(Configuration *conf) 
{
    array_destroy(conf->environment, environment_destroy);
    free(conf);
}


/**
 * \brief Initializes global options to default values.
 * \param conf Structure to initialize.
 */
void configuration_set_defaults(Configuration *conf)
{
    conf->max_cpuload = 0.0;
    conf->max_load_intv = 20;
    conf->uid = getuid();
    conf->gid = 0;
    conf->update_interval = 1.0;
    conf->local = 0;
    conf->port = 0;
    conf->unix_socket_permissions = 0660;
    conf->init_command[0] = 0;
    conf->authentication = AP_AUTH_NONE;
    conf->log_syslog_enabled = 1;
    conf->log_syslog_level = LOG_LEVEL_INFO;
    strcpy(conf->pid_file, DEFAULT_PIDFILE);
}


/**
 * \brief Appends an environment variable.
 * Appends an environment variable to a Configuration structure.
 * \param opt A Configuration.
 * \param attr Variable name.
 * \param value value.
 */
void configuration_add_environment(Configuration *opt, const char *attr, const char *value) 
{
    environment_array_add(opt->environment, attr, value);
}


/**
 * \brief Assigns a value to a field of a Configuration structure.
 * This function is use to assign field values to a Configuration structure.
 * The field name and the value are given as strings.
 * \param opt Global options.
 * \param attr Field name (upper case)
 * \param value Value
 */
void configuration_set_attribute(Configuration *opt, const char *attr, const char *value)
{
    if (!strcmp(attr, "MAX_CPULOAD"))
    {
	opt->max_cpuload = atof(value) * 0.01; /* argument is in percent */
	if ((opt->max_cpuload > 100.0) || (opt->max_cpuload < 0.0))
	{
	    opt->max_cpuload = 0.0;
	}
    }
    else if (!strcmp(attr, "MAX_LOAD_INTV"))
    {
	opt->max_load_intv = atoi(value);
	if (opt->max_load_intv < 1)
	{
	    opt->max_load_intv = 1;
	}
	else if (opt->max_load_intv > 1000)
	{
	    opt->max_load_intv = 1000;
	}
    }
    else if ( (!strcmp(attr, "UPDATE_INTERVAL"))
	      || (!strcmp(attr,"UPDATE"))
	      || (!strcmp(attr, "T_UPDATE")))
    {
	opt->update_interval = atof(value);
	if ((opt->update_interval < 0.01) && (opt->update_interval > 10.0))
	{
	    opt->update_interval = 1.0;
	}
    }
    else if (!strcmp(attr, "USER"))
    {
	if (!sys_util_findUID(value, &opt->uid, &opt->gid))
	{
	    opt->uid = -1;    /* Not found; set to NOBODY */
	    opt->gid = -1;
	}
    }
    else if (!strcmp(attr, "GROUP"))
    {
	if (!sys_util_findGID(value, &opt->gid))
	{
	    opt->gid = -1;
	}
    }
    else if (!strcmp(attr, "PORT"))
    {
	opt->port = atoi(value);
    }
    else if ((!strcmp(attr, "INIT_COMMAND"))
	     || (!strcmp(attr, "INITSCRIPT"))
	     || (!strcmp(attr, "INITCMD")))
    {
	snprintf(opt->init_command, OPTIONS_LEN_FNAME, "%s", value);
    }
    else if ((!strcmp(attr, "PIDFILE")))
    {
	snprintf(opt->pid_file, OPTIONS_LEN_FNAME, "%s", value);
    }
    else if ((!strcmp(attr, "UNIX_SOCKET_PERMISSIONS"))
	|| (!strcmp(attr, "LOCAL_PERM")))
    {
	sscanf(value,"%o", &(opt->unix_socket_permissions));
    }
    else if (!strcmp(attr, "AUTHENTICATION"))
    {
	if (!strcmp(value, "NONE"))
	{
	    opt->authentication = AP_AUTH_NONE;
	}
	else if (!strcmp(value, "CHANGE"))
	{
	    opt->authentication = AP_AUTH_CHANGE;
	}
	else if (!strcmp(value, "ALL"))
	{
	    opt->authentication = AP_AUTH_ALL;
	}
    }
    else if (!strcmp(attr, "LOG_SYSLOG_ENABLED") )
    {
	opt->log_syslog_enabled = string_util_bool_to_int(value);
    }
    /* REMOVED. log_console set in setup_initialize!
    else if (!strcmp(attr, "LOG_CONSOLE_ENABLED"))
    {
	opt->log_console = string_util_bool_to_int(value);
    }
    */
    else if (!strcmp(attr, "LOG_SYSLOG_LEVEL"))
    {
	opt->log_syslog_level = log_translate_string(value);
    }
    else
    {
	log_warning("Unknown option %s.", attr);
    }
	
}

