#include "shell_command_array.h"
#include <string.h>
#include <malloc.h>

#define SHELL_COMMAND_MIN_DIM 10

static void shell_command_array_increase(ShellCommandArray *s);

/**
 * \brief Creates a Shell Command array.
 * \return Pointer to a new Shell Command array or null if memory allocation failed.
 */
ShellCommandArray* shell_command_array_create() 
{
    ShellCommandArray *new_shell_command_array;
    new_shell_command_array = malloc(sizeof(ShellCommandArray));
    if (new_shell_command_array != NULL)
    {
	shell_command_array_init(new_shell_command_array);
    }
    return new_shell_command_array;
    
}

/** 
 * \brief Clear the contents of a script Array
 * Sets the array size to 0.
 * \param s Pointer to a script array
 */
void shell_command_array_clear(ShellCommandArray *s)
{
    s->dim = 0;
}


/** 
 * \brief Initializes a shell command array.
 * 
 * \param s Pointer to a shell command  array
 *
 * This function allocates memory for the array's contents
 * and initializes all variables. A certain amount of memory
 * is initially allocated as defined by the constant SHELL_COMMAND_MIN_DIM.
 * A call to shell_command_array_append() will automatically increase the
 * allocated memory if necessary.
 */
void shell_command_array_init(ShellCommandArray *s)
{
    s->data = (ShellCommand *) NULL;
    s->dim = 0;
    s->maxdim = 0;
    
    s->data = malloc(SHELL_COMMAND_MIN_DIM * sizeof(ShellCommand));
    if (s->data != NULL)
    {
	s->maxdim = SHELL_COMMAND_MIN_DIM;
    }
}


/**
 * Frees allocated memory of a shell command array.
 */
void shell_command_array_free(ShellCommandArray *s)
{
    free(s->data);
    free(s);
}


/** 
 * \brief Increase the memory size of a shell command array.
 * 
 * \param s Pointer to a shell command array
 * 
 * Used internally.
 */
static void shell_command_array_increase(ShellCommandArray *s)
{
    ShellCommand *newData;
    int newSize;
    
    // ** Kein Maxint-Check!
    newSize = 2 * s->maxdim;
    newData = (ShellCommand *) malloc(newSize * sizeof(ShellCommand));

    if (newData != NULL)
    {
	memcpy(newData, s->data, s->maxdim * sizeof(ShellCommand));
	free(s->data);
	s->data = newData;
	s->maxdim = newSize;
    }
}


/** 
 * \brief Appends a new data set to a shell command array.
 * 
 * ]param s Pointer to a shell command array
 * 
 * \return Pointer to the shell command data set or NULL on error
 * 
 * This function increases the dimension of a shell command array
 * and returns a pointer to the last data set. It allocates
 * memory automatically if necessary.
 */
ShellCommand *shell_command_array_append(ShellCommandArray *s)
{
    ShellCommand *rv = NULL;
    
    if (s->dim < s->maxdim)
    {
	rv = &(s->data[++s->dim]);
    }
    else
    {
	shell_command_array_increase(s);
	if (s->dim < s->maxdim)
	{
	    rv = &(s->data[++s->dim]);
	}
	else
	{
	    rv = NULL;
	}
    }
    shell_command_init(rv);
    return rv;
}


/** 
 * \brief Extracts an entry of a shell command array.
 * 
 * \param s Pointer to a shell command array
 * \param index Index of the desired entry
 * 
 * \return Pointer to a ShellCommand or NULL on error
 */
ShellCommand *shell_command_array_get(ShellCommandArray *s, int index)
{
    ShellCommand *rv;

    rv = NULL;
    if ((index > 0) && (index <= s->maxdim))
    {
	rv = &(s->data[index]);
    }
    return rv;
}


/** 
 * \brief Find a shell command data set by name.
 * 
 * \param s Pointer to a shell command array 
 * \param name 
 * 
 * \return Pointer to a ShellCommand or NULL on error.
 */
ShellCommand *shell_command_array_find(ShellCommandArray *s, const char *name)
{
    int found;
    ShellCommand *rv;
    int i;
    
    rv = NULL;
    
    found = 0;
    for(i=0; ((i<s->maxdim) && (!found)); i++)
    {
	if ((found = !strcmp(s->data[i].name, name)))
	{
	    rv = &(s->data[i]);
	}
    }
    return rv;
}
