#include "environment.h"
#include <string.h>
#include <malloc.h>
#include <stdlib.h>
#include "log.h"

/**
 * \brief Frees allocated memory for environment entry.
 */
void environment_destroy(void *data) 
{

    EnvironmentEntry *env;

    
    env = (EnvironmentEntry *)data;
    log_debug("environment_destroy %s", env->name);
    
    free(env->name);
    free(env->value);
    free(env);
}

/**
 * \brief Creates an environment entry.
 * \param name Name of the environment variable.
 * \param value Value
 */
EnvironmentEntry *environment_create(const char *name, const char *value) 
{
    EnvironmentEntry *newEntry = NULL;

    newEntry = (EnvironmentEntry *) malloc(sizeof(EnvironmentEntry));
    if (newEntry == NULL) {
	log_error("environment_create: Cannot allocate memory.");
    }
    else
    {
	if (!environment_init(newEntry, name, value))
	{
	    free(newEntry);
	    newEntry = NULL;
	}
    }
    return newEntry;
}



/**
 * \brief Initializes an environment entry.
 * \param env An environment entry.
 * \param name Name of the environment variable.
 * \param value Value
 */
int environment_init(EnvironmentEntry *env, const char *name, const char *value) 
{
    int len_name;
    int len_value;
    
    env->name = NULL;
    env->value = NULL;
    
    if (name != NULL) {
	len_name = strlen(name);
	env->name = malloc(sizeof(char) * (1+len_name));
	if (env->name != NULL) {
	    strcpy(env->name, name);
	} else {
	    log_warning("environment_create: Cannot allocate memory.");
	    return 0;
	}
    }
    if (value != NULL) {
	len_value = strlen(value);
	env->value = malloc(sizeof(char) * (1+len_value));
	if (env->value != NULL) {
	    strcpy(env->value, value);
	} else {
	    log_warning("environment_create: Cannot allocate memory.");
	    return 0;
	}
    }
    return 1;
}


/**
 * \brief Adds a new entry to an array of environment variables.
 * \param array An array.
 * \param attr Variable name.
 * \param value Value.
 */
void environment_array_add(Array *array, const char *attr, const char *value) 
{
    EnvironmentEntry *newEntry;
    newEntry = environment_create(attr, value);
    if (newEntry != NULL) {
	array_append(array, newEntry);
    }
}

/**
 * \brief Assigns environment variables as specified in an array of environment entries.
 * \param env Array of EnvironmentEntry.
 */
void environment_array_assign(Array *env) 
{
    int i;
    EnvironmentEntry *currentEnv;

    log_debug("environment_array_assign() %d", env->length);
    
    /* set process environment variables */
    for (i=0; i<env->length; i++)
    {
	currentEnv = array_get(env, i);
	if (currentEnv->value[0] == 0)
	{
	    log_debug("unsetenv %s", currentEnv->name);
	    unsetenv(currentEnv->name);
	}
	else
	{
	    log_debug("setenv %s = %s", currentEnv->name, currentEnv->value);
	    setenv(currentEnv->name, currentEnv->value, 1);
	}
    }
}

