#include "interpreter.h"
#include <sys/types.h>
#include <stdlib.h>
#include "log.h"
#include "shell_command.h"
#include "process.h"
#include <string.h> /* strncpy,strlen & co */
#include <ctype.h> /* isspace & co */
#include <stdio.h> /* sprintf */
#include <pth.h> /* pth_write */
#include <unistd.h>
#include "string_util.h" /* string_util_split_command */
#include "process_control.h"
#include "connection.h"
#include "setup.h" /* for setup_reinitialize */
#include "auth.h"

#define MSG_AUTH_REQUIRED "Authentication required.\n"
#define MSG_LOCKED "Locked by another client.\n"
#define END_OF_OUTPUT ".\n"  		/* Marks end of output when processing is done. */

/**
 *\brief Calls log_info with predefined message format.
 * This macro is only used to log incoming commands.
 */
#define interpreter_log_command(X) log_info("Command: %s", (X))

static void interpreter_process_line(char *input_buffer, ConnectionData *cd);

/**
 * \brief Interprets an input line and executes a command.
 * \param input_buffer Buffer containing the input line - must be zero-terminated.
 * \param cd Connection data - contains output file descriptor and authentication information.
 */
static void interpreter_process_line(char *input_buffer, ConnectionData *cd)
{

#define INPUT_COPY_SIZE 128

    int nArgs;
    int nTags;
    int i;

    ShellCommand *script;

    char *command_line;
    char *command;
    
    char* args[64];
    char **tags;
    char input_copy[INPUT_COPY_SIZE];
    int bytes;

    bytes = strlen(input_buffer);
    
    /* remove trailing spaces */
    for (i=bytes-1;i>=0;i--)
    {
	if (isspace(input_buffer[i]))
	    input_buffer[i] = 0;
	else
	    break;
    }

    string_util_strncpy(input_copy, input_buffer, INPUT_COPY_SIZE);
    
    if (input_buffer[0] == 0) {
	connection_print(cd, END_OF_OUTPUT);
	return;	/* Empty String */
    }
    
    nArgs = string_util_split_command(input_buffer, &command_line,args,10);
    if (nArgs <1)
    {
	connection_print(cd, END_OF_OUTPUT);
	return;
    }

    /* args[0] is the command, args[1..n] are the process tags. */
    /* Assign to variables for better readability below */
    command = args[0];
    nTags = nArgs-1;
    tags = &args[1];
    
    if (!connection_check_auth_level(cd, AP_AUTH_ALL))	/* Authentication necessary for everything? */
    {
	connection_print(cd, MSG_AUTH_REQUIRED);
	connection_print(cd, END_OF_OUTPUT);
	return;
    }
    
    if ( (!strcmp(command, "list"))
	 || (!strcmp(command, "ls"))
	 || (!strcmp(command, "ps")) )
    {
	process_control_print_list(cd->fd);
    }
    else if ( (!strcmp(command, "status"))
	      || (!strcmp(command, "st"))
	      || (!strcmp(command, "stat")))
    {
	process_control_print_statistics(cd->fd);
    }    else if (!strcmp(command, "stop"))
    {
	interpreter_log_command(input_copy);
	
	if (!connection_check_auth_level(cd, AP_AUTH_CHANGE))
	{
	    connection_print(cd, MSG_AUTH_REQUIRED);
	}
	else if (!process_control_acquire_mutex())
	{
	    pth_write(cd->fd, MSG_LOCKED, sizeof(MSG_LOCKED) - 1);
	}
	else
	{
	    if (nTags > 0)
	    {
		process_control_shutdown_processes(tags,nTags);
	    }
	    else
	    {
		process_control_shutdown_all_processes();
	    }
	    process_control_release_mutex();
	}
    }
    else if (!strcmp(command, "restart"))
    {
	interpreter_log_command(input_copy);
	if (!connection_check_auth_level(cd, AP_AUTH_CHANGE))
	{
	    connection_print(cd, MSG_AUTH_REQUIRED);
	}
	else if (!process_control_acquire_mutex())
	{
	    connection_print(cd,MSG_LOCKED);
	}
	else
	{
	    if (nTags > 0)
	    {
		process_control_shutdown_processes(tags,nTags);
		process_control_start_processes(tags,nTags,0);
	    }
	    else
	    {				/* Restart all processes */
		process_control_shutdown_all_processes();
		process_control_start_all_processes(0);
	    }
		
	    process_control_release_mutex();
	}
    }
    else if (!strcmp(command, "start"))
    {
	interpreter_log_command(input_copy);
	if (!connection_check_auth_level(cd, AP_AUTH_CHANGE))
	{
	    connection_print(cd, MSG_AUTH_REQUIRED);
	}
	else if (!process_control_acquire_mutex())
	{
	    pth_write(cd->fd, MSG_LOCKED, sizeof(MSG_LOCKED) - 1);
	}
	else
	{
	    
	    if (nTags > 0)
	    {
		process_control_start_processes(tags,nTags,0);
	    }
	    else				/* Start all processes */
	    {
		process_control_start_all_processes(0);
		
	    }
	    process_control_release_mutex();
	}
	
    }
    else if ((!strcmp(command, "init")) || (!strcmp(command, "reload")))
    {
	interpreter_log_command(input_copy);
	if (!connection_check_auth_level(cd, AP_AUTH_CHANGE))
	{
	    connection_print(cd, MSG_AUTH_REQUIRED);
	}
	else if (!process_control_acquire_mutex())
	{
	    connection_print(cd, MSG_LOCKED);
	}
	else
	{

	    if (nTags > 0)		/* Re-initialize single processes */
	    {
		process_control_shutdown_processes(tags,nTags);
		process_control_reset_process_statistics(tags, nTags);
		process_control_start_processes(tags,nTags,1);
	    }
	    else			/* Re-Initialize autoproc */
	    {
		process_control_shutdown_all_processes();
		process_control_reset_all_process_statistics();
		setup_reinitialize();
		setup_execute_global_init();
		process_control_start_all_processes(1);
	    }
	    process_control_release_mutex();
	}
    }
    else if (!strcmp(command, "exec"))
    {
	interpreter_log_command(input_copy);
	
	if (!connection_check_auth_level(cd, AP_AUTH_CHANGE))
	{
	    connection_print(cd, MSG_AUTH_REQUIRED);
	}
	else
	{
	    if (nTags > 0)
	    {
		for(i=0;i<nTags;i++)
		{
		    script = shell_command_array_find(setup_get_shell_commands(),tags[i]);
		    if (script != NULL)
		    {
			shell_command_execute(script);
		    }
		    else
		    {
			connection_print(cd, "Identifier not found.\n");
		    }
		}
	    }
	}
    }
    else if (!strcmp(command, "pw"))
    {
	if (nTags == 1)
	    cd->authOK = auth_verifyPasswd(tags[0]);
	else
	    cd->authOK = 0;

	if (!cd->authOK)
	{
	    connection_print(cd, "Access denied.\n");
	    log_warning( "Authentication Failure"); 
	}
    }
    else
    {
	connection_print(cd, "Unknown Command.\n");
    }
    
    /* END_OF_OUTPUT marks end of messages */
    connection_print(cd, END_OF_OUTPUT);
}




/**
 * \brief Processes input from a file descriptor.
 *
 * Reads from a file descriptor line by line and executes commands.
 * Also writes output to the file descriptor.
 *
 * \param fd - A file descriptor.
 *
 */
void interpreter_process_input(int fd)
{

#define INPUT_BUFFER_SIZE 4096

    char input_buffer[INPUT_BUFFER_SIZE];
    char *ptrBuffer;
    char *ptrLine;
    char *ptrNextLine;
    
    int bytes_left;
    int bytes_read;
    int rest_length;

    ConnectionData conn;

    connection_init(&conn);
    conn.fd = fd;

    bytes_left = INPUT_BUFFER_SIZE -1;
    ptrBuffer = input_buffer;
    ptrLine = input_buffer;
    rest_length = 0;

    log_debug("interpreter_process_input: begin %d", fd);

    
    /*
     * Read a number of bytes and divide them into lines.
     * Process each line separartely.
     */
    while ((bytes_read = pth_read(fd, ptrBuffer, bytes_left)) > 0)
    {

	ptrBuffer[bytes_read] = 0;	/* Needed for strchr & strlen below! */
	ptrLine = input_buffer;

	do
	{
	    ptrNextLine = strchr(ptrLine, '\n');
	    if (ptrNextLine != NULL)
	    {
		*ptrNextLine = 0;	/* Mark End of Line */
		interpreter_process_line(ptrLine, &conn);
		ptrLine = ptrNextLine + 1; /* Beginning of next Line */
	    }
	    else
	    {
		/*
		  if (strlen(ptrLine) > 0)
		  process_line(ptrLine, strlen(ptrLine), fd);
		*/
	    }

	} while (ptrNextLine != NULL);

	rest_length = strlen(ptrLine);

	bytes_left = INPUT_BUFFER_SIZE - rest_length - 1;
	if (rest_length > 0)
	{
	    memmove(input_buffer, ptrLine, rest_length * sizeof(char));
	    input_buffer[rest_length] = 0;
	}
	ptrBuffer = &input_buffer[rest_length]; /* Append new bytes here */
    }

    log_debug("interpreter_process_input: closing %d", fd);

    close(fd);
}

