#include <stdio.h>
#include <stdarg.h>
#include <syslog.h>
#include <stdlib.h>
#include <string.h>
#include "log.h"
#include <errno.h>
#include "string_util.h"

/**
 * \brief Entry in a list of log handlers
 */
typedef struct log_handler_list_entry
{
    LogHandler handler;			/**< Pointer handler function */
    int level;				/**< Log level for this handler */
    struct log_handler_list_entry *next;	/**< Pointer to next list entry */
} LogHandlerListEntry;


#define DOMAIN_LEN 64
#define DEFAULT_LEVEL LOG_LEVEL_INFO

static char m_domain[DOMAIN_LEN];

static LogHandlerListEntry *m_log_handlers = NULL;
static LogHandlerListEntry *log_create_handler_entry(LogHandler handler, int level) ;
static const char* print_to_buffer(const char *format, va_list args);


static LogHandlerListEntry *log_create_handler_entry(LogHandler handler, int level) 
{
    LogHandlerListEntry *result;
    result = (LogHandlerListEntry *)malloc(sizeof(LogHandlerListEntry));
    if (result != NULL) {
	result->handler = handler;
	result->next = NULL;
	result->level = level;
    }
    return result;
}

/**
 * \brief Adds a log handler
 */
void log_add_handler(LogHandler handler, int level) 
{
    LogHandlerListEntry *newEntry;
    LogHandlerListEntry *current;

    newEntry = log_create_handler_entry(handler, level);
    if (newEntry != NULL) {
	if (m_log_handlers == NULL) {
	    /* first list entry */
	    m_log_handlers = newEntry;
	}
	else
	{
	    /* go to end of list and append. */
	    current = m_log_handlers;
	    while (current->next != NULL)
		current = current->next;
	    current->next = newEntry;
	}
    }
}

/**
 * \brief Removes all log handlers.
 */
void log_remove_handlers() 
{
    LogHandlerListEntry *current;
    LogHandlerListEntry *next;
    
    current = m_log_handlers;

    while(current != NULL) {
	next = current->next;
	free(current);
	current = next;
    }
    m_log_handlers = NULL;
}

/**
 * \brief Initializes logging.
 */
void log_initialize(const char *domain) 
{
    string_util_strncpy(m_domain, domain, DOMAIN_LEN);
}

#define MSG_BUFFER_LEN 1024
static const char* print_to_buffer(const char *format, va_list args) 
{
    static char msg_buffer[MSG_BUFFER_LEN];
    vsnprintf(msg_buffer,  MSG_BUFFER_LEN, format, args);
    return msg_buffer;
}


/**
 * \brief Writes a log message.
 * \param level Level assigned to the message.
 * \param format Format string, like in printf().
 * \param args List of arguments
 */
void log_vwrite(int level, const char *format, va_list args)
{
    LogHandlerListEntry *current;
    const char *msg_buffer = NULL;
    
    current = m_log_handlers;
    while(current != NULL) {
	if (level >= current->level) {

	    if (msg_buffer == NULL) {
		msg_buffer = print_to_buffer(format, args);
	    }

	    current->handler(level, m_domain, msg_buffer);
	}
	current = current->next;
    }


}

/**
 * \brief Prints a log message with level LOG_LEVEL_INFO.
 */
void log_info(const char *format, ...) 
{
    va_list args;
    va_start(args, format);
    log_vwrite(LOG_LEVEL_INFO, format, args);
    va_end(args);
}

/**
 * \brief Prints a log message with level LOG_LEVEL_DEBUG.
 */
void log_debug(const char *format, ...) 
{
    va_list args;
    va_start(args, format);
    log_vwrite(LOG_LEVEL_DEBUG, format, args);
    va_end(args);
}

/**
 * \brief Prints a log message with level LOG_LEVEL_WARNING.
 */
void log_warning(const char *format, ...) 
{
    va_list args;
    va_start(args, format);
    log_vwrite(LOG_LEVEL_WARNING, format, args);
    va_end(args);
}

/**
 * \brief Perror equivalent.
 * Prints a log message which consists of a message text
 * and the description of the last error encountered,
 * separated by a colon.
 * \brief message Message text.
 * \see perror()
 */
void log_perror(const char *message) 
{
    log_error("%s: %s", message, strerror(errno));
}

/**
 * \brief Prints a log message with level LOG_LEVEL_ERROR.
 */
void log_error(const char *format, ...) 
{
    va_list args;
    va_start(args, format);
    log_vwrite(LOG_LEVEL_ERROR, format, args);
    va_end(args);
}

/**
 * \brief Prints a log message with level LOG_LEVEL_FATAL.
 */
void log_fatal(const char *format, ...) 
{
    va_list args;
    va_start(args, format);
    log_vwrite(LOG_LEVEL_FATAL, format, args);
    va_end(args);
}

/**
 * \brief Translates a string to a log level.
 * \param level_string Log level as a string.
 * \return For a string "X" or "x", LOG_LEVEL_X
 * is returned, where X is one of the log levels.
 * If the string is not recognized, LOG_LEVEL_ALL
 * is returned.
 */
int log_translate_string(const char *level_string) 
{
    int result;
    
    if ((!strcmp(level_string, "DEBUG"))
	|| (!strcmp(level_string, "debug")))
    {
	result = LOG_LEVEL_DEBUG;
    }
    else if ((!strcmp(level_string, "INFO"))
	     || (!strcmp(level_string, "info")))
    {
	result = LOG_LEVEL_INFO;
    }
    else if ((!strcmp(level_string, "WARNING"))
	     || (!strcmp(level_string, "warning")))
    {
	result = LOG_LEVEL_WARNING;
    }
    else if ((!strcmp(level_string, "ERROR"))
	     || (!strcmp(level_string, "error")))
    {
	result = LOG_LEVEL_ERROR;
    }
    else if ((!strcmp(level_string, "FATAL"))
	     || (!strcmp(level_string, "error")))
    {
	result = LOG_LEVEL_FATAL;
    }
    else
    {
	result = LOG_LEVEL_ANY;
    }
    return result;
}



/**
 * \brief Translates a log level to a string.
 */
const char *log_translate_level(int level) 
{
    const char *result;

    /* translate log level */
    switch(level) {
    case LOG_LEVEL_FATAL:
	result = "FATAL";
	break;
    case LOG_LEVEL_ERROR:
	result = "ERROR";
	break;
    case LOG_LEVEL_WARNING:
	result = "WARNING";
	break;
    case LOG_LEVEL_INFO:
	result = "INFO";
	break;
    case LOG_LEVEL_DEBUG:
	result = "DEBUG";
	break;
    default:
	result = "";
    }

    return result;
    
}


/**
 * \brief A log handler which writes to syslog.
 */
void log_syslog_handler(int level, const char *domain, const char *message)
{

    int syslog_level = 0;

    /* translate log level */
    switch(level) {
    case LOG_LEVEL_FATAL:
	syslog_level = LOG_CRIT;
	break;
    case LOG_LEVEL_ERROR:
	syslog_level = LOG_ERR;
	break;
    case LOG_LEVEL_WARNING:
	syslog_level = LOG_WARNING;
	break;
    case LOG_LEVEL_INFO: 
	syslog_level = LOG_INFO;
	break;
    case LOG_LEVEL_DEBUG:
	syslog_level = LOG_DEBUG;
	break;
    default:
	syslog_level = LOG_DEBUG;
    }
    
    /* print to syslog */
    syslog(syslog_level, "%s %s", log_translate_level(level), message);
    
}


/**
 * \brief A log handler which writes to the console.
 */
void log_console_handler(int level, const char *domain, const char *message)
{

    switch(level) {
    case LOG_LEVEL_FATAL:
    case LOG_LEVEL_ERROR:
    case LOG_LEVEL_WARNING:
	fprintf(stderr, "%s: %s %s\n", domain, log_translate_level(level), message);
	break;

    case LOG_LEVEL_INFO:
    case LOG_LEVEL_DEBUG:
	fprintf(stdout, "%s: %s %s\n", domain, log_translate_level(level), message);
	break;

    default:
	fprintf(stdout, "%s: %s\n", domain, message);
    }
    
}

