/* -*- C -*- */
#ifndef LOG_H
#define LOG_H

#include <stdarg.h>

/**
 * \brief The log levels.
 * Log levels in ascending priority.
 */
enum LogLevel {
    LOG_LEVEL_ANY = 0,			/**< Can be used as default level for log handlers: Below lowest priority */
    LOG_LEVEL_DEBUG = 1,		/**< For debugging messages - lowest priority  */
    LOG_LEVEL_INFO = 2,			/**< For informative messages  */
    LOG_LEVEL_WARNING = 3,		/**< For warnings  */
    LOG_LEVEL_ERROR = 4,		/**< For errors */
    LOG_LEVEL_FATAL = 5			/**< For fatal errors - highest priority.  */
};


/**
 * \brief Function prototype for a log handler
 * A log handler is responsible for processing a log message.
 * Log level filtering is done before the handlers are called. A log handler
 * will only be called if a log message's priority is equal or
 * higher than the log handler's priority. 
 * \param level The log level.
 * \param domain The domain.
 * \param message The log message.
 * \see log_console_handler()
 * \see log_syslog_handler()
 * \see log_add_handler()
 * \see log_translate_level()
 */
typedef void (* LogHandler) (int level, const char *domain, const char *message);

/**
 * \file log.h
 * \brief Autoproc logging.
 */
void log_initialize(const char *domain);
void log_remove_handlers();
void log_add_handler(LogHandler handler, int level);

void log_info(const char *format, ...);
void log_debug(const char *format, ...);
void log_error(const char *format, ...);
void log_perror(const char *message);
void log_fatal(const char *format, ...);
void log_warning(const char *format, ...);
void log_vwrite(int level, const char *format, va_list args);
const char *log_translate_level(int level) ;
int log_translate_string(const char *level_string);

void log_syslog_handler(int level, const char *domain, const char *message);
void log_console_handler(int level, const char *domain, const char *message);



#endif
