#include "process.h"
#include "environment.h"
#include "server.h" /* for server_close_sockets */

#include <stdio.h>
#include <dirent.h>
#include <signal.h>
#include <unistd.h>
#include "string_util.h"
#include "process.h"
#include "sys_util.h"
#include <string.h>
#include <stdlib.h>
#include <pth.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <array.h>
#include "log.h" /* also includes glib.h */

#define ARGS_MAX 32

#define PROCESS_CMD_MAX (PROCESS_LEN_COMMAND + PROCESS_LEN_ARGUMENTS)

#define DEFAULT_MAX_SUBSEQUENT_FAILURES 10
#define DEFAULT_FAILURE_TIMER 20

static int process_launch(ProcessData *proc,int init);
static int process_element_equals_tag(void *element, void *user_data);
static int process_element_equals_pid(void *element, void *user_data);

/**
 * \brief Executes a process.
 * \param proc Process data
 * \param execInitScript If != 0, execute the process's initialization script
 *                        before executing the process.
 * 
 * \return PID of the new child process if successful otherwise -1
 *
 * This function forks and spawns a child process which executes the
 * command given as an element of the process data. An initialization
 * script which is also specified in the process data can be executed
 * before the process.
 */
#define HOME_LEN 256
static int process_launch(ProcessData *proc, int execInitScript)
{
    char *cmd;
    int pid;
    int rv;
    int rc;
    
    char *args[ARGS_MAX];
    char cmdbuffer[PROCESS_CMD_MAX];
    char homedir[HOME_LEN];
    

    log_debug("process_launch %s %d", proc->tag, execInitScript);
    
    
    string_util_strncpy(cmdbuffer, proc->command, PROCESS_CMD_MAX);
    strncat(cmdbuffer, " ", PROCESS_CMD_MAX);
    strncat(cmdbuffer, proc->arguments, PROCESS_CMD_MAX);

    cmdbuffer[PROCESS_CMD_MAX -1] = 0;  /* cautiously set terminating 0 */
    string_util_split_command(cmdbuffer, &cmd, args, ARGS_MAX);

    //    pipe(proc->inputPipe);

    pid = pth_fork();
    if (pid < 0)
    {
	rv = -1;
    }
    else if (pid == 0) /* The child process */
    {

	/* server_close_sockets(); */
	atexit(NULL);
		
	/* change working directory */
	if(proc->wdir[0] != 0)
	{
	    if (chdir(proc->wdir) < 0)
		chdir("/");
	}

	/* Set UID and GID */
	if (getuid() == 0)
	{
	    if (proc->gid != 0)
	    {
		setgid(proc->gid);
	    }
	    if (proc->uid != 0)
	    {
		setuid(proc->uid);
		if (sys_util_findHomeDir(proc->uid, homedir, HOME_LEN))
		    setenv("HOME", homedir, 1);
      	    }
	}

	environment_array_assign(proc->environment);
	
	/* disconnect from stdin */
	proc->inputPipe[0] = open("/dev/null", O_RDONLY);
	dup2(proc->inputPipe[0],  fileno(stdin));

	proc->outputPipe[1] = open("/dev/null", O_WRONLY);
	
	/* connect stdout and stderr */
	dup2(proc->outputPipe[1], fileno(stdout));
	dup2(proc->outputPipe[1], fileno(stderr));
	
	/* Execute init script */
	rc = 1;
	if ((execInitScript) || (proc->init_on_restart))
	{
	    if (proc->init_command[0] != 0)
	    {
		rc = system(proc->init_command);
	    }
	    
	}

	/* Execute process */
	execvp(cmd, args);
	exit(EXIT_FAILURE);
    }
    else	       /* The parent process */
    {
	rv = pid;
    }
    return rv;
}


/**
 * \brief Checks if a process exists
 * \param proc Process data.
 * \return 1 if the process exists
 */
int process_exists(ProcessData *proc)
{
    /* Check if process exists by sending a 0 signal */ 
    return ((proc->pid > 0) && (!kill(proc->pid, 0)));
}


/**
 * \brief Sends a TERM signal to a process.
 * \param proc Process data.
 */
int process_terminate(ProcessData *proc)
{
    return kill(proc->pid, SIGTERM);
}


/**
 * \brief Kill a process.
 * \param proc Process data.
 */
int process_kill(ProcessData *proc)
{
    return kill(proc->pid, SIGKILL);
}


/*
 * \brief OLD VERSION Stops a process.
 * \param proc Process data.
 * \param msec_wait Wait time between TERM and KILL signal.
 *
 * This function sends a TERM signal to a process, waits for a
 * specified time, checks if the process has terminated, and
 * sends a KILL signal if the process still exists.
 * 
 */
/*
static int process_stop_old_version(ProcessData *proc, int msec_wait)
{
    int nSleep;

    nSleep = 0;

    log_info("Stopping %s", proc->tag);
    
    if ((proc->pid > 0) && process_exists(proc))
    {
	process_terminate(proc);
	
	while ((nSleep < 10) && process_exists(proc))
	{
	    nSleep++;
	    pth_usleep(100L * msec_wait);
	}
	
	if (process_exists(proc))
	    process_kill(proc);
    }
    proc->pid = 0;
    process_data_reset(proc);
    return 0;
}

*/

/**
 * \brief Stops a process.
 * \param proc Process data.
 * \param msec_wait Wait time between TERM and KILL signal.
 *
 * This function sends a TERM signal to a process, waits for a
 * specified time, checks if the process has terminated, and
 * sends a KILL signal if the process still exists.
 */
int process_stop(ProcessData *proc, int waitTimeMs)
{
    int nSleep;

    nSleep = 0;
    int waitTimeSum;

    log_debug("process_stop(proc=%s, waitTime=%d)", proc->tag, waitTimeMs);
   
    if ((proc->pid > 0) && process_exists(proc))
    {
	waitTimeSum = 0;

	/* Send TERM signal */
	log_info("Sending TERM signal to %s.", proc->tag);
	process_terminate(proc);
	
	while ((waitTimeSum < waitTimeMs) && process_exists(proc))
	{
	    pth_usleep(10000L); /* wait 10 ms */
	    waitTimeSum += 10;
	}
	
        /* if still alive, kill it */
	if (process_exists(proc))
	{
	    log_info("Sending KILL signal to %s after %d ms.", proc->tag, waitTimeSum);
	    process_kill(proc);
	}
	
    } else {
	log_info("process_stop: %s is not running.", proc->tag);
    }
    
    proc->pid = 0;
    process_data_reset(proc);
    return 0;
}

/**
 * \brief Stops a process.
 * \param proc Process data.
 * 
 * Behaves like process_stop(), but uses the
 * delay time as configured in proc->term_delay_ms .
 */
int process_shutdown(ProcessData *proc) 
{
    return process_stop(proc, (int)(proc->term_delay * 1000.0)); /* convert s to ms */
}



/**
 * \brief Starts a process.
 * \param proc Process data
 * \param execInitScript If true, execute an initialization script
 *                  before starting the process.
 * \return PID of the started process, or 0 if not successful.
 *
 * This function does not check if the specified process already exists.
 */
int process_start(ProcessData *proc,int execInitScript)
{

    log_info("Starting %s", proc->tag);
    proc->pid = process_launch(proc,execInitScript);
    process_notify_start(proc);
    
    usleep(100);
    
    if (!process_exists(proc))
    {
	proc->pid = 0;
    }
    return proc->pid;
}


/**  
 *  \brief Initializes process data with default values.
 */
void process_data_init(ProcessData *p, const char *tag)
{
    p->pid = 0;
    p->tag[0] = 0;
    p->state = ' ';
    p->cpuload = 0.0;
    p->command[0] = 0;
    p->required_state = 0;
    p->cputime = 0;
    p->wdir[0] = 0;
    p->description[0] = 0;
    p->start_time = 0;
    p->stop_time = 0;
    p->max_subsequent_failures = DEFAULT_MAX_SUBSEQUENT_FAILURES;
    p->uid = 0;				/* Don't change UID/GID */
    p->gid = 0;
    p->init_command[0] = 0;
    p->init_on_restart = 0;			/* Init on every start/restart */
    p->restart_policy = PROCESS_RESTART_YES;
    p->delay = 0.0;
    p->term_delay = 0.10;			/* default 100ms */
    process_reset_statistics(p);
    if (tag != NULL)
    {
	string_util_strncpy(p->tag, tag, PROCESS_LEN_TAG);
    }
    p->environment = array_create();
}


/**
 * \brief Delete old process information before restart.
 */
void process_data_reset(ProcessData *proc)
{
    proc->cpuload = 0.0;
    proc->state = 0;
    proc->cputime = 0;
    /* WRONG    array_clear(proc->environment, environment_destroy); */
}


/**
 * \brief Destroys a ProcessData structure and all its contents.
 * \param data Pointer to a ProcessData structure.
 */
void process_destroy(void *data) 
{
    ProcessData *proc;
    proc = (ProcessData *)data;
    array_destroy(proc->environment, environment_destroy);
    free(proc);
}


/**
 * \brief Resets the process statistics.
 * \param proc Process data.
 */
void process_reset_statistics(ProcessData *proc) 
{
    proc->n_starts = 0;
    proc->n_failures = 0;
    proc->n_truncations = 0;
    proc->n_subsequent_failures = 0;
    proc->failure_timer = 0;
}


/**
 * \brief Sets an attribute.
 * \param proc Pointer to Process data
 * \param attr Attribute's name
 * \param value Attribute's value
 */
void process_set_attribute(ProcessData *proc, const char*attr, const char *value) 
{
    char *env_name;
    char *env_value;
    static char bf_value[80];

    if ((!strcmp(attr,"ALIAS")) || (!strcmp(attr,"NAME")))
    {
	string_util_strncpy(proc->tag, value, PROCESS_LEN_TAG);
    }
    else if (!strcmp(attr,"ARGS"))
    {
	string_util_strncpy(proc->arguments, value, PROCESS_LEN_ARGUMENTS);
    }
    else if (!strcmp(attr, "PATH"))
    {
	string_util_strncpy(proc->command, value, PROCESS_LEN_COMMAND);
    }
    else if (!strcmp(attr, "USER"))
    {
	if (!sys_util_findUID(value, &proc->uid, &proc->gid))
	{
	    proc->uid = -1;    /* Not found; set to NOBODY */
	    proc->gid = -1;
	}
    }
    else if (!strcmp(attr, "GROUP"))
    {
	if (!sys_util_findGID(value, &proc->gid))
	{
	    proc->gid = -1;
	}
    }
    else if (!strcmp(attr, "ENV"))
    {
	/* NOTE : This is a relic from an earlier version. ENV should normally be a block! */
	string_util_strncpy(bf_value, value, 79);
	if (string_util_divide(bf_value, '=', &env_name, &env_value))
	{
	    process_add_environment(proc, env_name, env_value);
	}
    }
    else if ((!strcmp(attr, "ARGS")) || (!strcmp(attr, "ARG")))
    {
	string_util_cat(proc->arguments, " ", PROCESS_LEN_ARGUMENTS);
	string_util_cat(proc->arguments, value, PROCESS_LEN_ARGUMENTS);
    }
    else if (!strcmp(attr, "DESC"))
    {
	string_util_strncpy(proc->description, value, PROCESS_LEN_DESCRIPTION);
    }
    else if ((!strcmp(attr, "WPATH")) || (!strcmp(attr, "WDIR")))
    {
	string_util_strncpy(proc->wdir, value, PROCESS_LEN_WDIR);
    }
    else if (!strcmp(attr, "RESTART"))
    {
	if (!strcmp(value, "ALL"))
	    proc->restart_policy = PROCESS_RESTART_ALL;
	else if (!strcmp(value, "NO"))
	    proc->restart_policy = PROCESS_RESTART_NO;
	else
	    proc->restart_policy = PROCESS_RESTART_YES;
    }
    else if (!strcmp(attr, "MAX_RESTARTS")) /* Deprecated; replaced by MAX_SUBSEQUENT_FAILURES */
    {
	proc->max_subsequent_failures = atoi(value) - 1;
    }
    else if (!strcmp(attr, "MAX_SUBSEQUENT_FAILURES"))
    {
	proc->max_subsequent_failures = atoi(value);
    }
    else if (!strcmp(attr, "DELAY"))
    {
	proc->delay = atof(value);
	if (proc->delay <= 0.0)
	    proc->delay = 0.0;
	else if (proc->delay > 180.0)
	    proc->delay = 180.0;
    }
    else if (!strcmp(attr, "TERM_DELAY"))
    {
	proc->term_delay = atof(value);
	if (proc->term_delay <= 0.0) {
	    proc->term_delay = 0.0;
	} else if (proc->term_delay >= 100.0) {
	    proc->term_delay = 100.0;
	}
    }
    else if ((!strcmp(attr, "INIT_COMMAND"))
	     || (!strcmp(attr, "INITCMD")))
    {
	
	string_util_strncpy(proc->init_command, value, PROCESS_LEN_COMMAND);
    }
    else if (!strcmp(attr, "INIT_ON_RESTART"))
    {
	if (!strcmp(value, "YES"))
	    proc->init_on_restart = 1;
	else
	    proc->init_on_restart = 0;
    }
	
}


/**
 * \brief Adds an environment variable to the process data.
 * \param proc Process data.
 * \param attr Variable name.
 * \param value Value.
 */
void process_add_environment(ProcessData *proc,const char *attr, const char *value) 
{
    environment_array_add(proc->environment, attr, value);
}


/**
 * \brief Registers a failure.
 * This function has to be called when a process is to be restarted automatically.
 * It increments the total failure couter.
 * It increments the subsequent feilure couter if the failure timer is > 0;
 */
void process_notify_failure(ProcessData *proc) 
{
    proc->n_failures++;
    if (proc->failure_timer > 0)
    {
	proc->n_subsequent_failures++;
    } 
    proc->failure_timer = DEFAULT_FAILURE_TIMER;
}


/**
 * \brief Registers a process truncation.
 * \param proc Process data
 *
 * This function has to be called whenever a process has been stopped automatically.
 */
void process_notify_truncation(ProcessData *proc) 
{
    proc->n_truncations++;
}


/**
 * \brief Registers a process start.
 * This function is called by process_start.
 */
void process_notify_start(ProcessData *proc)
{
    proc->n_starts++;
}


/**
 * \brief Resets the failure counter.
 * \param proc Process data.
 *
 * This function has to be called whenever a process is started
 * intentionally. It resets the failure counter which
 * limits the number of automatic restarts.
 *
 * \see process_decrement_failure_timer() 
 */
void process_reset_failures(ProcessData *proc)
{
    proc->failure_timer = 0;
    proc->n_subsequent_failures = 0;
}


/**
 * \brief Decrements the failure timer.
 * \param proc Process data
 *
 * This function decrements the failure timer.
 * If the timer reaches 0, the counter of subsequent
 * failures is also set to 0.
 *
 * If a process crashes, the process control tries to restart it.
 * If a situation occurs where a process gets restarted over and over
 * again, the process control needs a criterion to detect this
 * as an error and give up restarting the process. This works with
 * a failure counter and a failure timer. Withe every restart of a process,
 * its failure timer is set to an initial value, and its counter of subsequent failures
 * is incremented. With every update interval, the failure timer is decremented.
 * When the failure timer reaches 0, the counter of subsequent failures is also
 * set to 0. When the counter of subsequent failures reaches a maximal value
 * before the failure timer reaches zero, then the process will not be restarted again.
 */
void process_decrement_failure_timer(ProcessData *proc) 
{
    if (proc->failure_timer > 0) {
	proc->failure_timer --;
    } else if (proc->n_subsequent_failures > 0) {
	log_debug("Resetting subsequent failure counter for %s", proc->tag);
	proc->n_subsequent_failures = 0;
    }
}


/**
 * \brief Returns true if the number of subsequent failures has reacehd its maximum.
 * \param proc Process data
 */
int process_has_max_subsequent_failures(ProcessData *proc) 
{
    return (proc->n_subsequent_failures >= proc->max_subsequent_failures);
}



/**
 * \brief Equals Function ony used by process_array_find_by_pid
 * \param element Pointer to a ProcessData struture
 * \param user_data Pointer to an int
 */
static int process_element_equals_pid(void *element, void *user_data) 
{
    int *pid_query;
    pid_query = (int *) user_data;
    
    return (((ProcessData *)element)->pid == *pid_query);
}


/**
 * \brief Fiends a process entry by PID.
 * \param array A process array
 * \param pid The process ID to look for
 * \return A pointer to a ProcessData structure or NULL
 * if no matching entry was found.
 */
ProcessData *process_array_find_by_pid(Array *array, int pid)
{

    return (ProcessData *) array_find_equals(array,process_element_equals_pid, (void *)(&pid));

}


/**
 * \brief Equals Function used by process_array_find_by_tag
 */
static int process_element_equals_tag(void *element, void *user_data) 
{
    return (!strcmp(((ProcessData *)element)->tag, (char *)user_data));
}


/**
 * \brief Fiends a process entry by tag.
 * \param array A process array
 * \param tag The process tag to look for
 * \return A pointer to a ProcessData structure or NULL
 * if no such matching entry was found.
 */
ProcessData *process_array_find_by_tag(Array *array, const char *tag)
{
    return (ProcessData *) array_find_equals(array,process_element_equals_tag, (void *)tag);
}







