/* -*-C-*- */
#ifndef PROCESS_H
#define PROCESS_H


/**
 * \file process.h
 * \brief Basic operations on processes.
 *
 * A process, its status and its environment are defined by
 * a ProcessData structure. The functions in this module
 * can start and stop processes described by that structure.
 * 
 */

#include <sys/timeb.h>
#include <sys/types.h>
#include "environment.h"
#include "array.h"


#define PROCESS_LEN_TAG 64		/**< Size of field tag  */
#define PROCESS_LEN_COMMAND 256		/**< Size of field command  */
#define PROCESS_LEN_ARGUMENTS 256	/**< Size of field arguments  */
#define PROCESS_LEN_WDIR 256		/**< Size of field wdir  */
#define PROCESS_LEN_DESCRIPTION 256	/**< Size of field description  */

/** Required state of a process */
enum RequiredState {
    PROCESS_ACTIVE,   /**< Process should be running */
    PROCESS_STANDBY   /**< Process should not be running */
};

/** Restart policy when a process has died */
enum RestartPolicy {
    PROCESS_RESTART_YES, /**< Restart this process */
    PROCESS_RESTART_NO,	/**< Ignore, don't restart  */
    PROCESS_RESTART_ALL	/**< Restart all processes  */
};

/** Actual Process State */
enum ProcessState {
    PROCESS_STATE_OFF,	/**< Process is not running */
    PROCESS_STATE_ON,	/**< Process is running  */
    PROCESS_STATE_ERR,	/**< Process could not be started successfully */
    PROCESS_STATE_WAIT 	/**< Process is currently being started */
};


/**
 * \brief Process information
 */
typedef struct {
    
    int pid;				/**< Process ID. */
    char tag[PROCESS_LEN_TAG];		/**< Unique identifier. */
    char state;				/**< Current state. */
    double cpuload;			/**< CPU load caused by this process. */
    char init_command[PROCESS_LEN_COMMAND];   /**< Shell command to execute before
					   (re)initialization. */
    char command[PROCESS_LEN_COMMAND]; /**< The path of the executable. */
    char arguments[PROCESS_LEN_ARGUMENTS]; /**< The command-line arguments. */
    
    int init_on_restart;		/**< Execute init script on start or restart? */
    int required_state;			/**< Requred state of this process.  */
    unsigned int max_subsequent_failures; /**< Maximal number of subsequent failures. */
    unsigned int cputime;		/**< Last value of utime + stime. */
    char wdir[PROCESS_LEN_WDIR];		/**< The working directory.  */
    char description[PROCESS_LEN_DESCRIPTION];	/**< A description.  */
    time_t start_time;			/**< Time when the process was started.  */
    time_t stop_time;			/**< Time when the process was stopped. */
    int restart_policy;			/**< Restart policy. */
    float delay;			/**< Delay after start in seconds. */
    float term_delay;		/**< Specified the minimal number of seconds autoprocd should wait before sending a KILL signal after a TERM signal to end the process. */
    uid_t uid;				/**< User ID. */
    gid_t gid;				/**< Group ID. */
    unsigned int n_starts;		/**< Number of starts in general. */
    unsigned int n_truncations;		/**< Number of automatic stops. */
    unsigned int n_failures;		/**< Number of all failures (crashes + truncations). */
    unsigned int n_subsequent_failures;	/**< Number of failures since last manual restart. */
    int failure_timer;                   /**< Number of update intervals within which a restart counts as subsequent. */
    Array *environment;		/**< Array of environment variables.  */
    
    int inputPipe[2];			/**< Pipe for stdin redirection.  */
    int outputPipe[2];			/**< Pipe for stdout redirection. */
    
} ProcessData;

void process_data_init(ProcessData *p, const char *tag);
void process_data_reset(ProcessData *pd);
int process_start(ProcessData *pd,int init); 
int process_stop(ProcessData *pd, int msec_wait);
int process_shutdown(ProcessData *pd);
int process_exists(ProcessData *pd);
int process_terminate(ProcessData *pid);
int process_kill(ProcessData *pid); 
void process_reset_statistics(ProcessData *pd);
void process_set_attribute(ProcessData *pd, const char*attr, const char *value);
void process_add_environment(ProcessData *pd, const char *attr, const char *value);
void process_notify_failure(ProcessData *pd) ;
void process_notify_truncation(ProcessData *pd) ;
void process_notify_start(ProcessData *pd);
void process_reset_failures(ProcessData *pd);
void process_decrement_failure_timer(ProcessData *pd);
int process_has_max_subsequent_failures(ProcessData *pd);
void process_destroy(void *data) ;

ProcessData *process_array_find_by_tag(Array *array, const char *tag);
ProcessData *process_array_find_by_pid(Array *array, int pid);

#endif

