#include <time.h>
#include <string.h>
#include <stdio.h>
#include <pth.h>
#include "log.h"
#include "process_info.h"
#include "cpu_info.h"
#include "process.h"
#include "configuration.h"
#include "process_control.h"
#include "shell_command.h"
#include "setup.h"


static Array *m_processes;
static ControlStatistics m_statistics;
static CpuInfo m_cpu_data;
static CpuInfo m_old_cpu_data;
static double m_sum_cpuload;
static int m_cpuload_cnt = 0;

/* Process control mutex. Must be acquired and released externally. */
static pth_mutex_t m_control_mutex;

static void process_control_init_statistics();
static void process_control_update_statistics();
static void process_control_update();
static void process_control_init_mutex();
static void process_control_watch_processes();


/**
 * \brief Print a process list to a file descriptor.
 * \param fd A file descriptor
 *
 * This function prints a process list in tabular form.
 * It is used for the "ls" interpreter command.
 */
void process_control_print_list(int fd)
{

    int i;
    char astr[6];
    char timebuffer[30];
    char outputBuffer[128];
    char current_state[10];

    time_t starttime;
    time_t now;
    
    struct tm *trv;
    struct tm tm_start;
    struct tm tm_now;

    ProcessData *proc;
    
    now = time(NULL);
    
    trv = localtime(&now);
    memcpy(&tm_now, trv, sizeof(tm_start));

    for (i=0;i<m_processes->length;i++)
    {
	proc = (ProcessData *)array_at(m_processes, i);
	if (proc == NULL) continue;
	
	if (proc->required_state == PROCESS_ACTIVE)
	    sprintf(astr, "ON");
	else
	    sprintf(astr, "OFF");


	starttime = m_cpu_data.btime + proc->start_time;
	trv = localtime(&starttime);
	memcpy(&tm_start, trv, sizeof(tm_start));

	while (tm_start.tm_year >= 100) tm_start.tm_year -= 100;
	sprintf(timebuffer, "%02d.%02d.%02d %02d:%02d",
		tm_start.tm_mday, tm_start.tm_mon + 1, tm_start.tm_year,
		tm_start.tm_hour, tm_start.tm_min);

	switch(proc->state)
	{
	case 'S':
	case 'R':
	case 'D':
	    strcpy(current_state, "ON");
	    break;
	default:
	    if (process_has_max_subsequent_failures(proc))
		strcpy(current_state, "ERR");
	    else
		strcpy(current_state, "OFF");
	    break;
	}
	
	sprintf(outputBuffer,"%-10s %s %3s %4s %5.1f %4d %4d \"%s\"\n",  
		proc->tag,
		timebuffer,
		astr,
		current_state,
		(double)(proc->cpuload * 100.0),
		proc->n_starts,
		proc->n_failures,
		proc->description
	    );

	pth_write(fd,outputBuffer,strlen(outputBuffer));
    }
}


/**
 * \brief Updates process and CPU data from the operating system.
 */
static void process_control_update()
{

    int i;
    unsigned long cpu_jiffies;
    unsigned long proc_jiffies;
    unsigned long sum_proc_jiffies;

    static ProcessInfo process_buffer;
    ProcessData *proc;
    
    m_old_cpu_data = m_cpu_data;
    sum_proc_jiffies = 0L;
    m_sum_cpuload = 0.0;
    
    if (cpu_info_update(&m_cpu_data))
    {
      
	cpu_jiffies = (m_cpu_data.utime - m_old_cpu_data.utime )
	    + (m_cpu_data.itime - m_old_cpu_data.itime)
	    + (m_cpu_data.stime - m_old_cpu_data.stime)
	    + (m_cpu_data.ntime - m_old_cpu_data.ntime);
	
	for(i=0;i<m_processes->length;i++)
	{
	    proc = (ProcessData *)array_at(m_processes, i);
	    if (proc == NULL) continue;
				     
	    if ((proc->pid > 0) && (process_info_update(proc->pid, &process_buffer)))
	    {
		if (proc->start_time == process_buffer.starttime_sec) 
		{

		    /* Process has not been restarted*/

		    proc_jiffies = process_buffer.cputime - proc->cputime;

		    if ((proc_jiffies < 0) || (proc_jiffies > cpu_jiffies))
			proc_jiffies = 0;
			
		    sum_proc_jiffies += proc_jiffies;
		    

		    if (cpu_jiffies > 0)
			proc->cpuload = (double)proc_jiffies/(double)cpu_jiffies;
		    else
			proc->cpuload = 0.0;
		    
		}
		else			/* Process has been restarted */
		{
		    proc->cpuload = 0.0;
		}

		proc->start_time = process_buffer.starttime_sec;
		proc->state = process_buffer.state;
		proc->cputime = process_buffer.cputime;

	    }
	    else			/* No process info available */
	    {
		proc->cpuload = 0.0;
		proc->state = 'N';
	    }
	}

	if (cpu_jiffies > 0)
	    m_sum_cpuload = (double)sum_proc_jiffies/(double)cpu_jiffies;

    }
}


/**
 * \brief Check process status and CPU load
 * and restarts or stops processes.
 */
static void process_control_watch_processes()
{
    int i;
    int restart_all;
    double avg_cpuload;
    double proc_delay;
    int restart_ok;
    Configuration *config;
    ProcessData *proc;
    
    
    if (!pth_mutex_acquire(&m_control_mutex, TRUE, NULL))
	return;
    
    restart_all = 0;
    proc_delay = 0;

    config = setup_get_configuration();

    for (i=0;i<m_processes->length;i++)
    {

	proc = (ProcessData *)array_at(m_processes, i);
	if (proc == NULL) continue;
	
	if (process_exists(proc))
	{
	    if (proc->required_state == PROCESS_STANDBY)	
	    {				/* Process should not be running */
		process_stop(proc,1);
		process_notify_truncation(proc);
		process_reset_failures(proc); /* cautiously */
	    } else {
		process_decrement_failure_timer(proc);
	    }
	}
	else				/* !process_exists() */
	{

	    /* The process is required to be active, but does not run. */
	    if (proc->required_state == PROCESS_ACTIVE)
	    {

		/* The process may have crashed or may have been truncated */
		/* Do not notify failures if max. number of failures has been reached. */
		if (!process_has_max_subsequent_failures(proc)) {
		    process_notify_failure(proc);
		    log_debug("process_control_watch_processes: failure %d of %s",
			      proc->n_subsequent_failures, proc->tag);
		}
		
		/* Restart until the maximum number of subsequent failures is exceeded. */
		if (!process_has_max_subsequent_failures(proc)) {

		    switch(proc->restart_policy)
		    {
		    case PROCESS_RESTART_ALL:
			if (restart_ok)
			    restart_all = 1;
			break;
		    case PROCESS_RESTART_NO:
			break;
		    case PROCESS_RESTART_YES:
		    default:
			if (restart_ok)
			{
			    if(proc_delay > 0.0)
			    {
				log_debug("process_control_watch_processes: delaying %d s", proc_delay);
				pth_usleep( (unsigned long)(proc_delay * 1.0E6));
			    }

			    process_start(proc,0);
			    proc_delay = proc->delay;
			}
			break;
		    }
		}
	    }
	}
    } /* for */

    /* Max. CPU load exceeded? */
    if ((config->max_cpuload > 0.0)
	&& ( m_sum_cpuload > config->max_cpuload))
    {
	m_cpuload_cnt++;
	log_warning("Maximal CPU load exceeded: %0.1f%%",
		 m_sum_cpuload * 100.0);
	
	if (m_cpuload_cnt > config->max_load_intv)
	{
	    m_cpuload_cnt = 0;
	    restart_all = 0;
	    
	    if (m_processes->length > 0)
		avg_cpuload = m_sum_cpuload / (double) (m_processes->length);
	    else
		avg_cpuload = 0.0;

	    /* If restart_all is set, all processes will be restarted.
	     * If at least one process is found consuming too much CPU time,
	     * this process will be stopped restarted depending on its restart policy.
	     */
	    for (i=0;(i<m_processes->length) && (!restart_all);i++)
	    {
		proc = (ProcessData *)array_at(m_processes, i);
		if (proc == NULL) continue;

		if (proc->cpuload >= avg_cpuload)
		{
		    if (proc->restart_policy == PROCESS_RESTART_ALL)
		    {
			restart_all = 1;
		    }
		    else
		    {
			log_warning("Stopping %s (%0.1f%% CPU load)",
				 proc->tag, proc->cpuload * 100.0);
			process_stop(proc,1); /* Will be restarted on next call */
			process_notify_truncation(proc);
		    }
		}
	    } /* stop all */
	} /* Max. cpuload count */
    }/* max. cpuload exceeded*/
    else
    {
	m_cpuload_cnt = 0;
    }
    
    if (restart_all)			/* restart all processes */
    {

	/* Note: This restart is not registered as a truncation */ 
	log_info("Restarting all Processes");
	process_control_shutdown_all_processes();
	process_control_start_all_processes(0);
	
    } /* restart all */

    pth_mutex_release(&m_control_mutex);

}


/**
 * \brief Process information update thread.
 *
 * Calls process_control_update() periodically.
 */
void *process_control_update_thread(void *arg)
{
    for (;;)
    {
	pth_usleep((unsigned int)(setup_get_configuration()->update_interval * 1.0e6));
	process_control_update();
	process_control_update_statistics();
	process_control_watch_processes();
    }
    return NULL;
}


/**
 * \brief Initializes the process control.
 *
 * \param procArray Process array.
 *
 * Assigns an array of processes to the process control.
 * Initializes mutex and statistics.
 *
 * \note pth_init() has to be called before this function!
 */
void process_control_initialize(Array *procArray) 
{
    m_cpuload_cnt = 0;
    m_processes = procArray;
    cpu_info_update(&m_cpu_data);
    process_control_init_mutex();
    process_control_init_statistics();
}


/**
 *  \brief Initilalize control mutex
 *  \internal
 */
static void process_control_init_mutex() 
{
    static int mutex_initialized = 0;

    if (!mutex_initialized)
    {
	pth_mutex_init(&m_control_mutex);
	mutex_initialized = 1;
    }
}


/**
 * \brief Releases the process control mutex
 */
int process_control_release_mutex()
{
    return pth_mutex_release(&m_control_mutex);
}


/**
 * \brief Acquire a mutex for process control.
 *
 * Before a process control function is called from an external module,
 * it is necessary to lock the process control. If one thread acquires a mutex,
 * it is locked by this thread and must be released to let other
 * processes access. The correct handling of locking/unlocking the
 * process control is left to the client. Process control functions
 * do not lock or unlock resources internally.
 * 
 * \return 1 if mutex could be acquired, 0 if mutex is locked.
 */
int process_control_acquire_mutex()
{
    return pth_mutex_acquire(&m_control_mutex, TRUE, NULL);
}


/**
 *  \brief Initializes or resets control statistics
 */
static void process_control_init_statistics()
{
    m_statistics.nActive  = 0;
    m_statistics.nErrors = 0;
    m_statistics.nRestarts = 0;
    m_statistics.nCrashes = 0;
    m_statistics.cpuload = 0.0;
}



/**
 * \brief Returns the process statistics.
 */
ControlStatistics *process_control_get_statistics() 
{
    return &m_statistics;
}


/**
 * \brief Update the process statistics
 */
void process_control_update_statistics()
{
    int i;
    ProcessData *proc;
    
    process_control_init_statistics();

    m_statistics.nProcesses = m_processes->length;
    
    for(i=0;i<m_processes->length;i++)
    {
	proc = (ProcessData *)array_at(m_processes, i);
	if (proc == NULL) continue;

	switch(proc->state)
	{
	case 'S':
	case 'R':
	case 'D':
	    m_statistics.nActive++;
	    break;
	default:
	    m_statistics.nInactive++;
	    if (process_has_max_subsequent_failures(proc)
		|| (proc->required_state == PROCESS_ACTIVE)) 
		m_statistics.nErrors++;
	    break;
	}
	m_statistics.nRestarts += (proc->n_starts -1);
	m_statistics.nCrashes += proc->n_failures;
	m_statistics.cpuload += proc->cpuload;
    }
}


/**
 * \brief Prints statistics output to a file descriptor, using pth_write.
 * \param fd File descriptor
 */
void process_control_print_statistics(int fd) 
{
    static char outputBuffer[256];
    sprintf(outputBuffer,
	    "Processes %5d\nActive %8d\nErrors %8d\nRestarts %6d\nCrashes %7d\nCPU-Load %6.1f\n",
	    m_statistics.nProcesses,
	    m_statistics.nActive,
	    m_statistics.nErrors,
	    m_statistics.nRestarts,
	    m_statistics.nCrashes,
	    m_statistics.cpuload * 100.0);

    pth_write(fd, outputBuffer, strlen(outputBuffer));
}


/**
 * \brief Starts a list of processes identified by their tags.
 * \param tags Pointer to a string array which contains the process tags.
 * \param nTags lenght of tags.
 * \param execInit Execute the process's init script before restart.
 */
void process_control_start_processes(char **tags, int nTags, int execInit) 
{
    ProcessData *proc;
    int i;

    for (i=0; i<nTags; i++) {

	proc = process_array_find_by_tag(m_processes,tags[i]);
	if (proc != NULL)
	{
	    if (!process_exists(proc))
	    {
		process_start(proc, execInit);
		process_reset_failures(proc);
		process_data_reset(proc);
		if ((proc->delay > 0.0) && (i<nTags-1))
		{
		    log_debug("process_control_start_processes: delaying %f s", proc->delay);
		    pth_usleep( (unsigned long)(proc->delay * 1.0E6));
		}
	    }
	    proc->required_state = PROCESS_ACTIVE;
	}
    }
}


/*
 * \brief DEPRECATED Stops a list of processes identified by their tags.
 * \param tags Pointer to a string array which contains the process tags.
 * \param nTags lenght of tags.
 */
/*
static void process_control_stop_processes_old_version(char **tags, int nTags) 
{
    ProcessData *proc;
    int i;

    for (i=0;i<nTags;i++) {
	proc = process_array_find_by_tag(m_processes,tags[i]);
	if (proc != NULL)
	{
	    process_stop(proc,1);
	    proc->required_state = PROCESS_STANDBY;
	}
    }
}
*/

/**
 * \brief Stops processes in reverse starting order.
 * \param tags Pointer to a string array which contains the process tags.
 * \param nTags lenght of tags.
 */
void process_control_shutdown_processes(char **tags, int nTags) 
{
    ProcessData *proc;
    int i;

    for (i=nTags-1;i>=0;i--) {
	proc = process_array_find_by_tag(m_processes,tags[i]);
	if (proc != NULL)
	{
	    process_shutdown(proc);
	    proc->required_state = PROCESS_STANDBY;
	}
    }
}


/**
 * \brief Starts all processes.
 *
 * \param execInit Execute the process's init script before restart.
 *
 * Starts all processes in the process array.
 */
void process_control_start_all_processes(int execInit) 
{
    ProcessData *proc;
    int i;

    /* Start all processes */
    for(i=0;i<m_processes->length;i++)
    {
	proc = (ProcessData *)array_at(m_processes, i);
	if (proc == NULL) continue;

	if (!process_exists(proc))
	{
	    process_start(proc,execInit);
	    process_reset_failures(proc);
	    process_data_reset(proc);
	    if ((proc->delay > 0.0) && (i<m_processes->length - 1))
	    {
		log_debug("process_control_start_all_processes: delaying %f s", proc->delay);
		pth_usleep( (unsigned long)(proc->delay * 1.0E6));
	    }
	}
	proc->required_state = PROCESS_ACTIVE;
    }
}


/**
 * \brief Stops all processes.
 *
 * Stops all processes in the process array.
 */
void process_control_stop_all_processes() 
{

    int i;
    ProcessData *proc;
    
    for(i=0;i<m_processes->length;i++)
    {
	proc = (ProcessData *)array_at(m_processes, i);
	if (proc == NULL) continue;

	process_stop(proc,1);
	proc->required_state = PROCESS_STANDBY;
    }
}

/**
 * \brief Stops all processes in reverse starting order.
 *
 * Stops all processes in the process array.
 * Sends a TERM signal, waits for the process to stop,
 * possibly sends a KILL signal after the process's
 * term_duration.
 * 
 */
void process_control_shutdown_all_processes() 
{

    int i;
    ProcessData *proc;

    /* stop processes in reverse order */
    for(i=m_processes->length-1;i>=0;i--)
    {
	proc = (ProcessData *)array_at(m_processes, i);
	if (proc == NULL) continue;

	process_shutdown(proc);
	proc->required_state = PROCESS_STANDBY;
    }
}



/**
 * \brief Resets the process statistics of a list of processes.
 * \param tags Pointer to a string array which contains the process tags.
 * \param nTags lenght of tags.
 */
void process_control_reset_process_statistics(char **tags, int nTags) 
{
    ProcessData *proc;
    int i;

    for (i=0;i<nTags;i++) {
	proc = process_array_find_by_tag(m_processes,tags[i]);
	if (proc != NULL)
	{
	    process_reset_statistics(proc);
	}
    }
}

/**
 *   \brief Resets the statistics of all processes.
 */
void process_control_reset_all_process_statistics() 
{
    ProcessData *proc;
    int i;

    /* stop processes */
    for(i=0;i<m_processes->length;i++)
    {
	proc = (ProcessData *)array_at(m_processes, i);
	if (proc == NULL) continue;
	process_reset_statistics(proc);
    }
}


/**
 *  \brief Sends a TERM signal to all processes.
 *  Sends a signal and does nothing else.
 *  For internal use.
 */
static void process_control_terminate_all() 
{
    ProcessData *proc;
    int i;

    /* stop processes */
    for(i=0;i<m_processes->length;i++)
    {
	proc = (ProcessData *)array_at(m_processes, i);
	if (proc == NULL) continue;
	if (process_exists(proc))
	{
	    process_terminate(proc);
	}
    }
}


/**
 *  \brief Sends a KILL signal to all processes.
 *  Sends a signal and does nothing else.
 *  For internal use.
 */
static void process_control_kill_all() 
{
    ProcessData *proc;
    int i;

    /* stop processes */
    for(i=0;i<m_processes->length;i++)
    {
	proc = (ProcessData *)array_at(m_processes, i);
	if (proc == NULL) continue;
	if (process_exists(proc))
	{
	    process_kill(proc);
	}
    }
}


/**
 *  \brief Emergency exit.
 *  Sends terminate and kill signals to all controlled processes.
 */
void process_control_close_all() 
{
    process_control_terminate_all();
    process_control_kill_all();
}


/**
 *  \brief Notifies the termination of a process.
 *
 *  When a process terminates, its parent process receives
 *  a CHILD signal. The signal handler calls this function
 *  to tell the process control that a process has terminated.
 */
void process_control_notify_termination(pid_t pid) 
{

    ProcessData *proc;
    time_t now;
    
    proc = process_array_find_by_pid(m_processes,pid);
    if (proc != NULL)
    {
	now = time(NULL);
	proc->start_time = now - m_cpu_data.btime;
	proc->pid = 0;
    }
}
