
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <arpa/inet.h>
#include <netdb.h>
#include <pth.h>
#include <tcpd.h>
#include <sys/un.h>
#include "server.h"
#include "interpreter.h"
#include "configuration.h"
#include "setup.h"
#include "sockets.h"
#include "log.h"

#define UNIX_SOCKET_NAME "/tmp/autoproc-0"
#define CONN_REFUSED "Connection refused.\n.\n"

static int m_inet_socket = 0;
static int m_unix_socket = 0;
static pth_mutex_t m_socket_mutex;
static void server_open_sockets();


/**
 * \brief Creates a new UNIX domain socket
 * \return File descriptor if successful, or error code
 * (<0) from the socket() call. 
 */
static int server_create_unix_socket (char *socketname)
{
    int m_unix_socket;
    
    /*sprintf(socketname, socketname); */
    /*    test_socket(socketname); */

    log_debug("server_create_unix_socket(%s)", socketname);

    unlink(socketname);
    m_unix_socket = sockets_unix_create(socketname);
    if (m_unix_socket < 0)
    {
	log_fatal("Creating IPC socket");
	exit(1);
    }
    else
	chmod(socketname, (mode_t) setup_get_configuration()->unix_socket_permissions);

    return m_unix_socket;
}


/**
 * \brief Check host-based access permission
 * for a client.
 * This functions uses tcp_wrappers to determine if
 * access is perimitted for a client.
 * \return 1 if access is permitted, else 0.
 */
static int server_is_access_permitted(struct sockaddr_in *sa)
{
    struct request_info rq;

    request_init(&rq, RQ_DAEMON, "autoprocd",
		 RQ_CLIENT_SIN, sa,
		 RQ_CLIENT_ADDR, inet_ntoa(sa->sin_addr),
		 0);
    
    return hosts_access(&rq);
}


/**
 * \brief TCP Handler thread
 * For each connection, a separate PTH thread is created.
 */
static void *server_tcp_connection_thread(void *arg)
{
    int socket_number;
    socket_number = *((int *)arg);
    log_debug("Starting interpreter on socket %d",socket_number);
    interpreter_process_input(socket_number);
    free(arg);
    return NULL;
}


/**
 * \brief Opens internet and unix socket.
 */
static void server_open_sockets() 
{

    Configuration *config;
    struct servent *port_servent;

    log_debug("server_open_sockets()");
    
    config = setup_get_configuration();
    
    if (getuid() == 0)			/* Read port from /etc/services */
    {
	port_servent = getservbyname("autoprocd", "tcp");
	if (port_servent != NULL)
	    config->port = port_servent->s_port;
    }

    /* Create socket  */
    if (m_unix_socket == 0)
    {
	m_unix_socket = server_create_unix_socket(UNIX_SOCKET_NAME);
	/* socket is not inherited to child processes */
	fcntl(m_unix_socket, F_SETFD, FD_CLOEXEC);
	if (listen(m_unix_socket, 1) < 0)
	{
	    /* TODO Improve error handling. Dont exit, just log. */
	    log_fatal("Listen (UNIX socket): %s", strerror(errno));
	    exit(-1);
	}
    }

    /* Create the TCP socket. */
    if (m_inet_socket == 0)
    {
	if (!config->local)
	{
	    m_inet_socket = sockets_inet_create(config->port);
	    /* socket is not inherited to child processes */
	    fcntl(m_inet_socket, F_SETFD, FD_CLOEXEC);
	    if (listen(m_inet_socket, 1) < 0)
	    {
		/* TODO Improve error handling. Dont exit, just log. */
		server_shutdown();
		log_fatal("Listen (INET socket): %s", strerror(errno));
		exit(-1);
	    }
	}
    }
}


/**
 * \brief Closes server sockets.
 *
 * This function is called when the server shuts down or
 * when a new process is forked.
 */
void server_close_sockets() 
{
    if (m_unix_socket) close(m_unix_socket);
    if (m_inet_socket) close(m_inet_socket);
    m_unix_socket = 0;
    m_inet_socket = 0;
}


/**
 * \brief Server shtdown.
 *
 * Closes all sockets, deletes the pipe.
 */
void server_shutdown() 
{
    log_debug("server_shutdown()");
    server_close_sockets();
    unlink(UNIX_SOCKET_NAME);
}


/**
 * \brief Server main function.
 *
 * Waits for connecting clients, spawns connection threads.
 */
void *server_main_thread(void *arg)
{

    fd_set read_fd_set;
    int permissionOK;
    int *new_socket;	/* dynamically allocated thread data after each accept; deleted by thread. */
    Configuration *config;
    int rc;
    pth_attr_t thread_attr;
    struct timeval socket_timeout;
    
    struct sockaddr_un un_client;
    struct sockaddr_in in_client;
    size_t sockaddr_size;

    config = setup_get_configuration();

    /* Thread_Attributes for all connection threads */
    thread_attr = pth_attr_new();
    pth_attr_set(thread_attr, PTH_ATTR_JOINABLE, PTH_PRIO_STD);

    pth_mutex_init(&m_socket_mutex);
    
    socket_timeout.tv_sec = 5;
    socket_timeout.tv_usec = 0;

    /* Main Thread: Listen to sockets */ 
    server_open_sockets();
    new_socket = NULL;
    while (1)
    {

	log_debug("Acquiring Mutex");
	pth_mutex_acquire(&m_socket_mutex, FALSE, NULL); /* protects local variables */

	FD_ZERO(&read_fd_set);
	if (m_unix_socket > 0) FD_SET(m_unix_socket, &read_fd_set);
	if (m_inet_socket > 0) FD_SET(m_inet_socket, &read_fd_set);

	errno = 0;
	log_debug("Selecting socket");
	rc = pth_select(FD_SETSIZE, &read_fd_set, NULL, NULL, NULL);
	if ((rc >= 0)  && (errno == 0)) /* */	 /* No error (e.g. signal) */
	{
	    log_debug("Select OK %d", rc);

	    if (FD_ISSET(m_unix_socket, &read_fd_set))
	    {
		sockaddr_size = sizeof(un_client);
		log_info("pth_accept");
		
		new_socket = (int *)malloc(sizeof(int)); /* The socket must be allocated dynamically, since it
							  * is passed as an argument to the server thread.
							  * The thread is responsible for the deletion.
							  */ 
		if (new_socket != NULL)
		{
		    *new_socket = pth_accept(m_unix_socket,
					     (struct sockaddr *) &un_client,
					     &sockaddr_size);
		    log_debug("accepted %d", *new_socket);
		
		    if (*new_socket > 0) 
		    {
			/* new_socket is not inherited to child processes */
			fcntl(*new_socket, F_SETFD, FD_CLOEXEC);
			log_info("Local connection"); 
			pth_spawn(thread_attr, server_tcp_connection_thread, new_socket);
			new_socket = NULL;
		    }
		    else
		    {
			free(new_socket);
			log_perror("Accept (UNIX socket)");
		    }
		    
		}
		else
		{
		    log_perror("Malloc (UNIX socket)");
		}
	    }
	    else if ((!config->local) && (FD_ISSET(m_inet_socket, &read_fd_set)))
	    {
		
		sockaddr_size = sizeof(in_client);

		new_socket = (int *)malloc(sizeof(int)); /* The socket must be allocated dynamically, since it
							  * is passed as an argument to the server thread.
							  * The thread is responsible for the deletion.
							  */ 
		if (new_socket != NULL)
		{
		    *new_socket = pth_accept(m_inet_socket, (struct sockaddr *)
					&in_client, &sockaddr_size);
		    if (*new_socket > 0)
		    {

			/* check if client is allowed to connect */
			permissionOK = server_is_access_permitted(&in_client);

			if (permissionOK)
			{
			    /* new_socket is closed on exec in child processes */
			    fcntl(*new_socket, F_SETFD, FD_CLOEXEC);
			    log_info("Network connection from %s",
				     inet_ntoa(in_client.sin_addr)); 
			    if (new_socket > 0)
				pth_spawn(thread_attr, server_tcp_connection_thread, new_socket);
			}
			else
			{
			    log_info("Network connection refused: %s",
				     inet_ntoa(in_client.sin_addr)); 
			    pth_write(*new_socket, CONN_REFUSED, sizeof(CONN_REFUSED));
			    close(*new_socket);
			    free(new_socket);
			}
		    }
		    else
		    {
			free(new_socket);
			log_perror("Accept(Network connection)");
		    }
		} /* socket allocated */
		else
		{
		    log_perror("Malloc (Network connection");
		}
		
	    }  /* ISSET Network socket */
	} /* socket >= 0 */
	else
	{
	    /* Frequent error messages: Resource temporarily unavailable; don't print as error. */
	    log_debug("Select error: %d %d %s", rc, errno, strerror(errno));
	}
	
	log_debug("Releasing mutex");
	pth_mutex_release(&m_socket_mutex);
	
    } /* while(1) */
    
    log_debug("autoprocd: shutting down");
    server_shutdown();
	
}


