#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h> /* openlog */
#include "string_util.h"
#include "sys_util.h"
#include "shell_command.h"
#include "configuration.h"
#include "connection.h" /* auth_modes */
#include "malloc.h"
#include "config_parser.h"
#include "setup.h"
#include "log.h"
#include "process.h"
#include "array.h"
#include "options.h"

static Configuration *m_configuration;
static int m_inside_script = 0;
static int m_inside_proc = 0;
static int m_inside_env = 0;
static Array *m_processes;
static Array *m_shell_commands;
static ShellCommand *m_current_script;
static ProcessData *m_current_proc;
static Options m_command_line_options;

static char *m_config_file;

static void setup_read_config(const char *inputFile);
static void setup_init_defaults();
static void setup_handler_set_attr(const char *attr, const char *value);
static void setup_handler_begin_block(const char *type, const char *name);
static void setup_handler_end_block();
static void setup_assign_environment(Configuration *opt);
static void setup_install_log_handlers();

/**
 * \brief Begin block handler for config_parser.
 */
static void setup_handler_begin_block(const char *type, const char *name)
{
    ProcessData *newProcess;
    
    if ((type != NULL) && (!m_inside_proc) && (!m_inside_script)) 
    {
	if (!strcmp(type, "PROCESS"))
	{
	    m_inside_proc = 1;

	    newProcess = (ProcessData *)malloc(sizeof(ProcessData));
	    if (newProcess != 0) {
		
		array_append(m_processes, newProcess);
		process_data_init(newProcess, name);
		m_current_proc = newProcess;

	    } else {
		log_error("Memory allocation failed for process %s!", name);
		m_current_proc = NULL;
	    }
	}
	else if (!strcmp(type, "SHELLCMD"))
	{
	    m_inside_script = 1;
	    m_current_script = shell_command_create();
	    if (m_current_script != NULL)
	    {
		array_append(m_shell_commands, m_current_script);
		if (name != NULL)
		{
		    string_util_strncpy(m_current_script->name, name, SHELL_COMMAND_LEN_NAME-1);
		}		
	    }
	}
	else if (!strcmp(type, "ENV"))
	{
	    m_inside_env = 1;
	} else {
	    /* TODO invalid block ! */
	}
	
	
    } /* else it's "GLOBAL" */
    else
    {
	if (!strcmp(type, "ENV"))
	{
	    m_inside_env = 1;
	} else {
	    /* TODO invalid block! */
	}
	
    }
}


/**
 * \brief End block handler for config_parser.
 */
static void setup_handler_end_block()
{
    if (m_inside_env)
    {
	m_inside_env = 0;
    }
    else if (m_inside_proc)
    {
	m_current_proc = NULL;
	m_inside_proc = 0;
    }
    else if (m_inside_script)
    {
	m_current_script = NULL;
	m_inside_script = 0;
    }
}


/**
 * \brief Attribute handler for config_parser.
 */
static void setup_handler_set_attr(const char *attr, const char *value)
{

    if ((attr == NULL) || (value == NULL)) return;

    if (m_inside_env)			/* Inside ENV block */
    {
	if (m_current_proc != NULL) /* Process Environment */
	{
	    process_add_environment(m_current_proc, attr, value);
	}
	else if (m_current_script != NULL)
	{
	    shell_command_add_environment(m_current_script, attr, value);
	}
	else /* Global Environment */
	{
	    configuration_add_environment(m_configuration, attr, value);
	}
    }
    else if ((m_inside_proc) && (m_current_proc != NULL))
    {
	process_set_attribute(m_current_proc, attr, value);
	
    } /* inside block && current_proc*/
    else if (m_inside_script && (m_current_script != NULL))
    {					/* Shell command block */
	shell_command_set_attribute(m_current_script, attr, value);
    }
    else  /* if (m_opt != NULL) */
    { /* Global parameters. Not necessarily inside a block */
	configuration_set_attribute(m_configuration, attr, value);
    }
}


/**
 * \brief Reads the configuration file.
 * Sets options and reads process and shell command entries.
 * For internal use only. Called after initialization.
 * \param inputFile Configuration file.
 */
static void setup_read_config(const char *inputFile)
{

    m_inside_env = 0;
    m_current_proc = NULL;
    m_current_script = NULL;

    log_debug("Reading configuration file %s", inputFile);

    setup_init_defaults();
    
    if (m_processes != NULL)
    {
	config_parser_read(inputFile,
		 setup_handler_begin_block,
		 setup_handler_end_block,
		 setup_handler_set_attr);
    }
    
    m_configuration->local = (m_configuration->port == 0);	/* Don't create a network socket */

    openlog("autoproc",0,LOG_DAEMON);

}


/**
 * \brief Installs log handlers according to
 * configuration and command-line options.
 */
static void setup_install_log_handlers()
{
    log_remove_handlers();
    if (m_command_line_options.set_debug_enabled) 
    {
	log_add_handler(log_console_handler, m_command_line_options.debug_level);
    }
    if (m_configuration->log_syslog_enabled)
    {
	log_add_handler(log_syslog_handler, m_configuration->log_syslog_level);
    }
}


/**
 * \brief Sets option defauls.
 * This method is called before the configuration file is read.
 */
static void setup_init_defaults() 
{
    configuration_set_defaults(m_configuration);
}


/**
 * \brief Re-Initializes the setup.
 *
 * Re-initializes the process array and the shell command array,
 * re-reads the configuration file, 
 * resets the global options, assigns global environment
 * variables.
 */
void setup_reinitialize() 
{

    log_debug("setup_reinitialize()");
    
     /*
       renew syslog settings - not necessary
       closelog();
       openlog("autoprocd", 0, LOG_DAEMON);
       openlog("autoprocd", LOG_PERROR, LOG_DAEMON);
     */
    
    array_clear(m_processes, process_destroy);
    array_clear(m_shell_commands, shell_command_destroy);

    /* For m_configuration, clear the environment only! */
    array_clear(m_configuration->environment, environment_destroy);
    setup_read_config(m_config_file);
    
    setup_install_log_handlers();
    setup_assign_environment(m_configuration);
    
}




/**
 * \brief Executes the global initialization command
 * specified in configuration file.
 */
void setup_execute_global_init()
{
    int rc;
    Configuration *config;

    config = setup_get_configuration();
    
    if (config->init_command[0] != 0)
    {
	log_info("Executing initialization command.");
	rc = system(config->init_command);
	if (rc > 0)
	    log_error("Error in %s.", config->init_command);
	else if (rc < 0)
	    log_error("Can't execute %s.", config->init_command);
    }
}


/**
 * \brief Returns the process array.
 */
Array *setup_get_processes() 
{
    return m_processes;
}


/**
 * \brief Returns the shell command array.
 */
Array *setup_get_shell_commands() 
{
    return m_shell_commands;
}


/**
 * \brief Returns the global configuration parameters.
 */
Configuration *setup_get_configuration() 
{
    return m_configuration;
}


/**
 * \brief Creates and initializes global variables.
 * \param options Command-line options.
 * \param configFile Configuration file.
 * 
 * Creates a process array and a shell command array,
 * reads the configuration file, initializes everything,
 * sets the global options, assigns global environment
 * variables.
 *
 * \return Pointer to a Configuration.
 *
 */
Configuration *setup_initialize(Options *options, const char *configFile) 
{

    int slen;
    
    slen = strlen(configFile) + 1;
    m_config_file = (char *)malloc(slen * sizeof(char));
    m_shell_commands = array_create();
    m_processes = array_create();
    
    if ((m_processes == NULL) || (m_shell_commands == NULL) || (m_config_file == NULL))
    {
	log_fatal("setup_initialize: Out of Memory");
	exit(1);
    }

    string_util_strncpy(m_config_file, configFile, slen);
    m_configuration = configuration_create();

    /* Save command line options */
    m_command_line_options = *options;
    
    setup_read_config(m_config_file);
    setup_install_log_handlers();
    setup_assign_environment(m_configuration);

    return m_configuration;
    
}


/**
 * \brief Assigns environment variables as specified.
 *
 * This function calls setenv to assign environment
 * variables from the environment list in a Configuration
 * structure.
 *
 * \param opt Pointer to a configuration.
 */
static void setup_assign_environment(Configuration *opt) 
{
    environment_array_assign(opt->environment);
}



