#include <stdlib.h>
#include <string.h>
#include "log.h"
#include "shell_command.h"
#include <pth.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "sys_util.h"
#include "configuration.h"
#include "setup.h"
#include "environment.h"
#include "server.h"
#include "array.h"
#include "string_util.h"

extern void close_descriptors();


/**
 * \brief Creates and initializes a ShellCommand.
 */
ShellCommand *shell_command_create() 
{
    ShellCommand *result;
    Array *newEnv;

    result = (ShellCommand *)malloc(sizeof(ShellCommand));
    if (result == NULL) {
	log_error("shell_command_create: Cannot allocate memory!");
	return NULL;
    }

    newEnv = array_create();
    if (newEnv == NULL) {
	log_error("shell_command_create: Cannot allocate memory!");
	free(result);
	return NULL;
    }

    result->environment = newEnv;
    shell_command_set_defaults(result);
    return result;

}

/**
 * \brief Destroys a ShellCommand.
 * \param data Pointer to a ShellCommand.
 */
void shell_command_destroy(void *data) 
{
    ShellCommand *sc;
    sc = (ShellCommand *)data;
    
    array_destroy(sc->environment, environment_destroy);
    free(sc);
}

/**
 * \brief Searches an entry in an array of shell commands.
 * \param array GPtrArray of ShellCommands.
 * \param name Name to search for.
 * \return First entry with matching name or NULL if no matching entry found.
 */
ShellCommand *shell_command_array_find(Array *array, const char *name) 
{
    int i;
    ShellCommand *result = NULL, *current = NULL;

    for (i=0;i<array->length;i++)
    {
	current = (ShellCommand *)array_get(array, i);
	if (current != NULL) {
	    if (!strcmp(current->name, name))
	    {
		result = current;
		break;
	    }
	}
    }
   return result;
}



/**
 * \brief Initialize shell command data with default values.
 */
void shell_command_set_defaults(ShellCommand *command)
{
    if (command != NULL)
    {
	command->name[0] = 0;
	command->cmd[0] = 0;
	command->wpath[0] = 0;
	command->uid = -1;
	command->gid = -1;
    }
}

#define ARGS_MAX 32
#define CMD_MAX 256
#define HOME_LEN 256

/** 
 * \brief Executes a shell command in the background.
 * 
 * \param command Pointer to a ShellCommand 
 * 
 * \return 0 on success
 */
int shell_command_execute(ShellCommand *command)
{
    int pid;
    int rv;
    int rc;
    int fd0, fd1;
    char homedir[HOME_LEN];
    
    pid = pth_fork();
    if (pid < 0)
    {
	rv = -1;
    }
    else if (pid == 0) /* The child process */
    {

	/* server_close_sockets(); */
	atexit(NULL);
		
	/* change working directory */
	if(command->wpath[0] != 0)
	{
	    if (chdir(command->wpath) < 0)
		chdir("/");
	}

	/* Set UID and GID */
	if (getuid() == 0)
	{
	    if (command->gid != 0)
	    {
		setgid(command->gid);
	    }
	    if (command->uid != 0)
	    {
		setuid(command->uid);
		if (sys_util_findHomeDir(command->uid, homedir, HOME_LEN))
		    setenv("HOME", homedir, 1);
		
	    }
	}

	/* assign environment variables */
	environment_array_assign(command->environment);
	
	/* disconnect from stdin */
	fd0 = open("/dev/null", O_RDONLY);
	dup2(fd0,  fileno(stdin));

	fd1 = open("/dev/null", O_WRONLY);
	
	/* connect stdout and stderr */
	dup2(fd1, fileno(stdout));
	dup2(fd1, fileno(stderr));
	
	/* Execute init script */
	rc = 1;
	if (command->cmd[0] != 0)
	    rc = system(command->cmd);

	exit(EXIT_FAILURE);
    }
    else	       /* The parent process */
    {
	rv = pid;
    }
    return rv;

}


/**
 * \brief Sets an attribute
 *
 * This function can be used to assign values to a ShellCommand
 * structure. It is called from setup routines only.
 * \param command Pointer to a ShellCommand structure
 * \param attr Attribute's name (upper case) 
 * \param value Attribute's value
 */
void shell_command_set_attribute(ShellCommand *command, const char *attr, const char *value) 
{
    if (!strcmp(attr, "PATH"))
    {
	string_util_strncpy(command->cmd, value, SHELL_COMMAND_LEN_CMD);
    }
    else if (!strcmp(attr, "USER"))
    {
	if (!sys_util_findUID(value, &command->uid, &command->gid))
	{
	    command->uid = -1;    /* Not found; set to NOBODY */
	    command->gid = -1;
	}
    }
    else if (!strcmp(attr, "GROUP"))
    {
	if (!sys_util_findGID(value, &command->gid))
	{
	    command->gid = -1;
	}
    }
    else if ((!strcmp(attr, "WPATH")) || (!strcmp(attr, "WDIR")))
    {
	string_util_strncpy(command->wpath, value, SHELL_COMMAND_LEN_WPATH);
    }

}


/**
 * \brief Adds an environment variable.
 * \param command Shell command
 * \param attr Name of the variable
 * \param value Value
 */
void shell_command_add_environment(ShellCommand *command,const char *attr, const char *value) 
{
    environment_array_add(command->environment, attr, value);
}
