#include "sockets.h"

#include <stddef.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include "log.h"
#include "string_util.h"

/**
 * \brief Initializes a sockaddr_in structure
 * Assigns family, port and host address.
 * \param name Socked address
 * \param hostname host name
 * \param port port
 */
static int sockets_init_sockaddr (struct sockaddr_in *name,
	       const char *hostname,
	       uint16_t port)
{
    struct hostent *hostinfo;
     
    name->sin_family = AF_INET;
    name->sin_port = htons (port);
    hostinfo = gethostbyname (hostname);
    if (hostinfo == NULL)
    {
	log_error ("Unknown host %s.", hostname);
	return 0;
    }
    name->sin_addr = *(struct in_addr *) hostinfo->h_addr;
    return 1;
}


/**
 * \brief Creates and binds an internet socket.
 * Creates and prepares a TCP socket on which
 * a server can listen.
 * \param port Port number.
 */
int sockets_inet_create (int port)
{
    int sock;
    struct sockaddr_in name;
     
    /* Create the socket. */
    sock = socket (PF_INET, SOCK_STREAM, 0);
    if (sock < 0)
    {
	log_error("socket_inet_create: %s", strerror(errno));
	return -1;
    }
     
    /* Give the socket a name. */
    name.sin_family = AF_INET;
    name.sin_port = htons (port);
    name.sin_addr.s_addr = htonl (INADDR_ANY);
    if (bind (sock, (struct sockaddr *) &name, sizeof (name)) < 0)
    {
	log_error("socket_inet_create: %s", strerror(errno));
	return -1;
    }
     
    return sock;
}


/**
 * \brief Creates and initializes a unix domain socket.
 * \param filename Name of the pipe.
 * \return File descriptor.
 */
int sockets_unix_create (const char *filename)
{
    struct sockaddr_un name;
    int sock;
    size_t size;
     
    /* Create the socket. */
    sock = socket (PF_LOCAL, SOCK_STREAM, 0);
    if (sock < 0)
    {
	log_error("socket_unix_create: %s", strerror(errno));
	return -1;
    }
     
    /* Bind a name to the socket. */
    name.sun_family = AF_LOCAL;
    string_util_strncpy (name.sun_path, filename, sizeof (name.sun_path));
     
    /* The size of the address is
       the offset of the start of the filename,
       plus its length,
       plus one for the terminating null byte.
       Alternatively you can just do:
       size = SUN_LEN (&name);
    */
    size = (offsetof (struct sockaddr_un, sun_path)
	    + strlen (name.sun_path) + 1);
     
    if (bind (sock, (struct sockaddr *) &name, size) < 0)
    {
	log_error("socket_unix_create: %s", strerror(errno));
	close(sock);
	return -1;
    }
     
    return sock;
}


/**
 * \brief Creates a socket and connects to a remote host.
 * This function is used by the autoproc client.
 * \param hostname Host name.
 * \param port Port number.
 * \return File descriptor.
  */
int sockets_inet_connect(const char *hostname, int port)
{
    int sock;
    struct sockaddr_in servername;
     
    /* Create the socket. */
    sock = socket (PF_INET, SOCK_STREAM, 0);
    if (sock < 0)
	return -1;
     
    /* Connect to the server. */
    if (sockets_init_sockaddr (&servername, hostname, port))
    {
	if (0 > connect (sock,
			 (struct sockaddr *) &servername,
			 sizeof (servername)))
	{
	    close(sock);
	    return -1;
	}
	
    } else {
	close(sock);
	return -1;
    }

    return sock;
}



/**
 * \brief Creates a socket and connects to a FIFO.
 * \param filename Name of the FIFO.
 * \param opt fcntl options.
 * \return 
 */
int sockets_unix_connect (const char *filename, int opt)
{
    struct sockaddr_un name;
    int sock;
    size_t size;
    int oldflags;
     
    /* Create the socket. */
    sock = socket (PF_LOCAL, SOCK_STREAM, 0);
    if (sock < 0)
    {
	return -2;
    }

    oldflags = fcntl(sock, F_GETFL, 0);
    oldflags |= opt;
    fcntl(sock, F_SETFL, oldflags);

    /* Bind a name to the socket. */
    name.sun_family = AF_LOCAL;
    string_util_strncpy (name.sun_path, filename, sizeof (name.sun_path));
     
    /* The size of the address is
       the offset of the start of the filename,
       plus its length,
       plus one for the terminating null byte.
       Alternatively you can just do:
       size = SUN_LEN (&name);
    */
    size = (offsetof (struct sockaddr_un, sun_path)
	    + strlen (name.sun_path) + 1);
     
    if (connect(sock, (struct sockaddr *) &name, size) < 0)
    {
	close(sock);
	return -1;
    }
     
    return sock;

}


