#define _GNU_SOURCE			/* for strnlen */
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "string_util.h"


/**
 * \brief Remove leading and trailing whitespaces from a string.
 * \param txt A Text String.
 */
int string_util_trim(char *txt)
{
    int l;
    int i;
    int firstchar;

    /* find position of first non-whitespace character */
    l = strlen(txt);
    firstchar = 0;
    for (i=0;((i<l) && (txt[i] != 0));i++)
    {
	if (!isspace(txt[i]))
	{
	    firstchar = i;
	    break;
	}
    }

    /* strip trailing whitespaces */
    for(i=l-1;i>0;i--)
    {
	if (isspace(txt[i]))
	{
	    txt[i] = 0;
	    l--;
	}
	else
	    break;
    }

    /* In case of leading whitespaces, shift array left */
    if (firstchar != 0)
    {
	l -= firstchar;
	/* TODO replace by memcpy? */
	for(i=0;i<l;i++)
	    txt[i] = txt[i+firstchar];
	txt[l] = 0;
    }

    return l;

}


/*!
  \brief Split a command line into an argument list.

  This function splits a command line buffer into commands and arguments.
  If a number of subsequent whitespaces occurs, the first whitespace
  (end of substring) will be replaced by 0, and a pointer to the
  first following non-whitespace character (beginning of substring)
  will be stored in args[].

  The function returns an array of arguments and a pointer to a command
  string as required by execv & friends. These pointers are addresses
  in the specified command line buffer.

  \warning Make sure that *cmdline contains a copy of the original command
  line, since this string will be modified. 
  
  \param cmdline Buffer comtaining the command line - must be zero-terminated!
  \param cmd Pointer to a pointer where the filename is stored 
  \param args Array to store char pointers
  \param mxargs Size  of args[]

  \return Number of arguments
*/
int string_util_split_command(char *cmdline, char **cmd, char **args, int mxargs)
{
    int i;
    int nArgs;
    int space;
    int len;
    int quote;
    char quote_char;
    
    mxargs--;				/* reserve last index for NULL */
    space = 1;
    nArgs = 0;

    quote = 0;
    quote_char = '\'';

    len = strlen(cmdline);
    string_util_fast_trim(cmdline, &cmdline, 0);

    i = 0;
    while((cmdline[i]!=0)&&(i<len))
    {
	while (isspace(cmdline[i]) && (i<len)) /* Replace first whitespace by 0 */
	{
	    cmdline[i] = 0;
	    space = 1;
	    i++;
	}

	if ((cmdline[i] == '\"') || (cmdline[i] == '\''))
	{
	    quote_char = cmdline[i];
	    if ((i<len - 1))
	    {
		i++;
		if (nArgs < mxargs)
		{
		    args[nArgs] = &cmdline[i];
		    nArgs++;
		    args[nArgs] = NULL;
		}
		while((cmdline[i] != quote_char) && (i<len)) i++;

		if (cmdline[i] == quote_char)
		{
			cmdline[i] = 0;
			space = 1;
			i++;
		}
	    }
	}
	else
	{
	    if ((space) && (nArgs < mxargs))
	    {
		args[nArgs] = &cmdline[i];
		nArgs++;
		args[nArgs] = NULL;
		space = 0;
	    }
	    i++;
	}
    }

    if (args[0] != NULL)
    {
	/* Extract filename from first argument */
	*cmd = args[0];
	len = strlen(args[0]);
	if (len > 0)
	{
	    /* Find last slash  */
	    for (i=len-1;((i>0) && (args[0][i] != '/'));i--);
	    
	    if (i > 0)			/* Last slash at position i  */
	    {
		args[0] = &args[0][i+1];	/* args[0] := basename */
	    }
	}
    }

    return nArgs;
}

/**
 * \brief Remove leading and trailing whitespaces.
 * \param txt Input string.
 * \param dest A pointer where to store the result.
 * \param remove_quote If != 0, also remove enclosing double quote characters.
 * In opposite to string_util_trim(), this function
 * does not left-shift the string, but returns a pointer to the
 * first non-whitespace character.
 * \return Length of the resulting string.
 */
int string_util_fast_trim(char* txt, char **dest, int remove_quote)
{
    int len;
    int i;
    int imax;
    char *result;

    len = strlen(txt);

    /* strip trailing spaces and \n */
    imax = len -1;
    for(i=imax;((i>=0) && isspace(txt[i]));i--)
    {
	txt[i] = 0;
	len--;
    }
    

    /* strip leating spaces. */
    result = &txt[0];
    imax = len;
    for(i= 0;((i<imax) && isspace(txt[i]));i++)
    {
	result++;
	len--;
    }

    /* strip double quotes if the string begins and ends with them */
    if (remove_quote && (len > 1))
    {
	if ((result[0] == '"') && (result[len-1] == '"')) {
	    result[len-1] = 0;
	    result++;
	    len -= 2;
	}
    }

    *dest = result;
    return len;
}

/**
 * \brief Appends a source string to a destination string.
 * \param dest Destination string.
 * \param src Source string.
 * \param maxlen Maximal length of the destination string.
 */
void string_util_cat(char *dest, const char *src, int maxlen)
{
    int dlen;
    
    dlen = strlen(dest);

    if (dlen >= maxlen-1)
	return;
    else
	strncat(dest,src,maxlen-dlen-1);
}


/**
 * \brief Divides a string into two parts.
 * Divide a string buffer into two parts, separated by
 * a separator character. The content of the string
 * is changed by this function!
 *
 * \param buffer The string to be separated.
 * \param separator The separator character.
 * \param first Pointer to the first string.
 * \param second Pointer to the second string.
 *
 * \return 1 if successful, 0 if separator not found
 */
int string_util_divide(char *buffer, char separator, char **first, char **second)
{
    int rv;
    char *arg1, *arg2;


    arg1 = NULL;
    arg2 = NULL;
    
    arg2 = strchr(buffer, separator);
    if (arg2 != NULL)
    {
	/* replace colon with 0 and separate first and second part*/
	*arg2 = 0;
	arg2++;
	string_util_trim(arg2);

	arg1 = &buffer[0];
	string_util_trim(arg1);

	rv = 1;
    }
    else
    {
	rv = 0;
    }

    *first = arg1;
    *second = arg2;
    
    return rv;
}

/**
 * \brief converts a true/false string to an int.
 * \param txt input text.
 * \return Returns 1 if the input string equals "TRUE", "true", or "True",
 * and 0 otherwise. 
*/
int string_util_bool_to_int(const char *txt) 
{
    return ( (!strcmp(txt,"true")) || (!strcmp(txt, "TRUE")) || (!strcmp(txt, "True")) );
}


/**
 * \brief A safe wrapper for strncpy.
 * Guarantees that the result is zero-terminated.
 * \param dest Destination.
 * \param src Source.
 * \param n Maxmial length.
 */
char *string_util_strncpy(char *dest, const char * src, size_t n) 
{
    char *result;
    result = strncpy(dest, src, n);
    dest[n-1] = 0;
    return result;
}
